#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
increments.py

@author : Louis RICHARD
"""

import numpy as np
import xarray as xr

from scipy.stats import kurtosis


def increments(x=None, scale=10):
    """Returns the increments of a time series.

    .. math:: y = |x_i - x_{i+s}|

    where :math:`s` is the scale.

    Parameters
    ----------
    x : xarray.DataArray
        Input time series.

    scale : int
        Scale at which to compute the increments.

    Returns
    -------
    kurt : numpy.ndarray
        kurtosis of the increments, one per product, using the Fisher's
        definition (0 value for a normal distribution).

    result : xarray.DataArray
        An xarray containing the time series increments, one per
        product in the original time series.

    """
    assert x is not None and isinstance(x, xr.DataArray)
    assert isinstance(scale, int)

    data = x.data

    if len(data.shape) == 1:
        data = data[:, np.newaxis]
    else:
        pass

    f = np.abs((data[scale:, :] - data[:-scale, :]))

    result = np.array(f)

    t, c = x.dims

    time = x.coords[t].data
    cols = x.coords[c].data

    result = xr.DataArray(result, coords=[time[0:len(f)], cols], dims=[t, c], attrs=x.attrs)

    kurt = kurtosis(result, axis=0, fisher=False)

    return kurt, result



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
get_dist.py

@author : Louis RICHARD
"""

import bisect
import numpy as np

from spacepy import pycdf
from dateutil import parser

from ..pyrf import ts_skymap


def get_dist(file_path="", cdf_name="", tint=None):
    """Read field named cdf_name in file and convert to velocity distribution function.

    Parameters
    ----------
    file_path : str
        Path of the cdf file.

    cdf_name : str
        Name of the target variable in the cdf file.

    tint : list
        Time interval.

    Returns
    -------
    out : xarray.Dataset
        Time series of the velocity distribution function if the target specie in the selected time
        interval.

    """

    tmmode = cdf_name.split("_")[-1]
    with pycdf.CDF(file_path) as f:
        if tmmode == "brst":

            DEPEND_0 = f[cdf_name].attrs["DEPEND_0"]
            DEPEND_1 = f[cdf_name].attrs["DEPEND_1"]
            DEPEND_2 = f[cdf_name].attrs["DEPEND_2"]
            DEPEND_3 = f[cdf_name].attrs["DEPEND_3"]

            t = f[DEPEND_0][...]
            idx_left = bisect.bisect_left(t, parser.parse(tint[0]))
            idx_right = bisect.bisect_left(t, parser.parse(tint[1]))
            t = t[idx_left:idx_right]
            if not t.size:
                return None

            dist = f[cdf_name][idx_left:idx_right, ...]
            dist = np.transpose(dist, [0, 3, 1, 2])
            ph = f[DEPEND_1][idx_left:idx_right, ...]
            th = f[DEPEND_2][...]
            en = f[DEPEND_3][idx_left:idx_right, ...]

            en0_name = "_".join([cdf_name.split("_")[0], cdf_name.split("_")[1], "energy0",
                                 cdf_name.split("_")[-1]])
            en1_name = "_".join([cdf_name.split("_")[0], cdf_name.split("_")[1], "energy1",
                                 cdf_name.split("_")[-1]])
            d_en_name = "_".join([cdf_name.split("_")[0], cdf_name.split("_")[1], "energy_delta",
                                  cdf_name.split("_")[-1]])
            e_step_table_name = "_".join([cdf_name.split("_")[0], cdf_name.split("_")[1],
                                         "steptable_parity", cdf_name.split("_")[-1]])

            step_table = f[e_step_table_name][idx_left:idx_right, ...]
            if d_en_name in f.keys():
                delta_plus_var = f[d_en_name][idx_left:idx_right, ...]
                delta_minus_var = f[d_en_name][idx_left:idx_right, ...]

            if en0_name not in f.keys():
                energy0 = en[1, :]
                energy1 = en[0, :]
            else:
                energy0 = f[en0_name][...]
                energy1 = f[en1_name][...]

            res = ts_skymap(t, dist, None, ph, th, energy0=energy0, energy1=energy1,
                            esteptable=step_table)

            if "delta_plus_var" in locals() and "delta_minus_var" in locals():
                res.attrs["delta_energy_minus"] = delta_minus_var
                res.attrs["delta_energy_plus"] = delta_plus_var

            for k in f[cdf_name].attrs:
                res.attrs[k] = f[cdf_name].attrs[k]

            for k in f.attrs:
                res.attrs[k] = f.attrs[k]

            res.attrs["tmmode"] = tmmode
            if "_dis_" in cdf_name:
                res.attrs["species"] = "ions"
            else:
                res.attrs["species"] = "electrons"

        elif tmmode == "fast":
            DEPEND_0 = f[cdf_name].attrs["DEPEND_0"]
            DEPEND_1 = f[cdf_name].attrs["DEPEND_1"]
            DEPEND_2 = f[cdf_name].attrs["DEPEND_2"]
            DEPEND_3 = f[cdf_name].attrs["DEPEND_3"]
            t = f[DEPEND_0][...]
            idx_left = bisect.bisect_left(t, parser.parse(tint[0]))
            idx_right = bisect.bisect_left(t, parser.parse(tint[1]))
            t = t[idx_left:idx_right]
            dist = f[cdf_name][idx_left:idx_right, ...]
            dist = np.transpose(dist, [0, 3, 1, 2])
            ph = f[DEPEND_1][...]
            th = f[DEPEND_2][...]
            en = f[DEPEND_3][idx_left:idx_right, ...]
            res = ts_skymap(t, dist, en, ph, th)

            for k in f[cdf_name].attrs:
                res.attrs[k] = f[cdf_name].attrs[k]

            for k in f.attrs:
                res.attrs[k] = f.attrs[k]

            res.attrs["tmmode"] = tmmode
            if "_dis_" in cdf_name:
                res.attrs["species"] = "ions"
            else:
                res.attrs["species"] = "electrons"
    return res



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
dpflux_to_vdf.py

@author : Louis RICHARD
"""


import numpy as np
import xarray as xr

from astropy import constants

from .spectr_to_dataset import spectr_to_dataset


def dpflux_to_vdf(dpflux=None):
    """Compute density particle flux from velocity distribution function.

    Parameters
    ----------
    dpflux : xarray.Dataset or xarray.DataArray
        Time series of the density particle flux in 1/(cm^2 s sr keV).

    Returns
    -------
    vdf : xarray.Dataset
        Time series of the velocity distribution function. Units must be either s^3/m^6.

    """

    assert dpflux is not None and isinstance(dpflux, (xr.Dataset, xr.DataArray))

    if isinstance(dpflux, xr.DataArray):
        dpflux = spectr_to_dataset(dpflux)

    if dpflux.attrs["species"] in ["ions", "i"]:
        mm = 1
    elif dpflux.attrs["species"] in ["electrons", "e"]:
        mm = constants.m_e.value / constants.m_p.value
    else:
        raise ValueError("Invalid specie")

    if dpflux.attrs["UNITS"] == "1/(cm^2 s sr keV)":
        tmp_data = dpflux.data.data / 1e12 * 0.53707 * mm ** 2
    else:
        raise ValueError("Invalid unit")

    energy = dpflux.energy.data

    if tmp_data.ndim == 2:
        tmp_data = tmp_data[:, :, None, None]

    data_r = np.reshape(tmp_data,
                        (tmp_data.shape[0], tmp_data.shape[1], np.prod(tmp_data.shape[2:])))

    if energy.ndim == 1:
        energy_mat = np.tile(energy, (len(dpflux.time), np.prod(tmp_data.shape[2:]), 1))
        energy_mat = np.transpose(energy_mat, [0, 2, 1])
    elif energy.ndim == 2:
        energy_mat = np.tile(energy, (np.prod(tmp_data.shape[2:]), 1, 1))
        energy_mat = np.transpose(energy_mat, [1, 2, 0])
    else:
        raise ValueError("Invalid energy shape")

    data_r /= energy_mat
    tmp_data = np.reshape(data_r, tmp_data.shape)

    dpflux = dpflux.copy()
    dpflux.data.data = np.squeeze(tmp_data)
    dpflux.attrs["UNITS"] = "s^3/m^6"

    return dpflux



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
get_feeps_energy_table.py

@author : Louis RICHARD
"""

import numpy as np


def get_feeps_energy_table(mms_id="1", eye="top", sensor_id=1):
    """This function returns the energy table based on each spacecraft and eye; based on the
    table from : FlatFieldResults_V3.xlsx
    
    Parameters
    ----------
    mms_id : str
        Spacecraft index e.g., "4" for MMS4.

    eye : str
        Sensor eye #.

    sensor_id : int
        Sensor ID.

    Returns
    -------
    out : list
        Energy table.
        
    Notes
    -----
    Bad eyes are replaced by NaNs.
    Different original energy tables are used depending on if the sensor head is 6-8 (ions) or not
    (electrons) :

        * Electron Eyes: 1, 2, 3, 4, 5, 9, 10, 11, 12
        * Ion Eyes: 6, 7, 8

    """

    assert isinstance(mms_id, str) and mms_id in [str(i) for i in range(1, 5)]
    assert isinstance(eye, str) and eye in ["bot", "top"]
    assert isinstance(sensor_id, int) and sensor_id in np.arange(1, 13)

    table = {"mms1-top": [14., 7., 16., 14., 14., 0., 0., 0., 14., 14., 17., 15.],
             "mms1-bot": [np.nan, 14., 14., 13., 14., 0., 0., 0., 14., 14., -25., 14.],
             "mms2-top": [-1., 6., -2., -1., np.nan, 0., np.nan, 0., 4., -1., -1., 0.],
             "mms2-bot": [-2., -1., -2., 0., -2., 15., np.nan, 15., -1., -2., -1., -3.],
             "mms3-top": [-3., np.nan, 2., -1., -5., 0., 0., 0., -3., -1., -3., np.nan],
             "mms3-bot": [-7., np.nan, -5., -6., np.nan, 0., 0., 12., 0., -2., -3., -3.],
             "mms4-top": [np.nan, np.nan, -2., -5., -5., 0., np.nan, 0., -1., -3., -6., -6.],
             "mms4-bot": [-8., np.nan, -2., np.nan, np.nan, -8., 0., 0., -2., np.nan, np.nan, -4.]}

    if 6 <= sensor_id <= 8:
        mms_energies = [57.90, 76.80, 95.40, 114.1, 133.0, 153.7, 177.6, 205.1, 236.7, 273.2, 315.4,
                        363.8, 419.7, 484.2, 558.6, 609.9]
    else:
        mms_energies = [33.20, 51.90, 70.60, 89.40, 107.1, 125.2, 146.5, 171.3, 200.2, 234.0, 273.4,
                        319.4, 373.2, 436.0, 509.2, 575.8]

    out = [energy + table[f"mms{mms_id}-{eye}"][sensor_id - 1] for energy in mms_energies]

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
pres_anis.py

@author : Louis RICHARD
"""

import numpy as np
import xarray as xr
from astropy import constants

from .resample import resample
from ..mms import rotate_tensor


def pres_anis(p_xyz=None, b_xyz=None):
    """
    Compute pressure anisotropy factor:

    .. math::

        \\mu_0 \\frac{P_\\parallel - P_\\perp}{|\\mathbf{B}|^2}

    Parameters
    ----------
    p_xyz : xarray.DataArray
        Time series of the pressure tensor.

    b_xyz : xarray.DataArray
        Time series of the background magnetic field.

    Returns
    -------
    p_anis : xarray.DataArray
        Time series of the pressure anisotropy.

    See also
    --------
    pyrfu.mms.rotate_tensor : Rotates pressure or temperature tensor into another coordinate system.

    Examples
    --------
    >>> from pyrfu import mms, pyrf
    >>> # Time interval
    >>> tint = ["2015-10-30T05:15:20.000", "2015-10-30T05:16:20.000"]
    >>> # Spacecraft index
    >>> mms_id = 1
    >>> # Load magnetic field, ion/electron temperature and number density
    >>> b_xyz = mms.get_data("B_gse_fgm_srvy_l2", tint, mms_id)
    >>> p_xyz_i = mms.get_data("Pi_gse_fpi_fast_l2", tint, mms_id)
    >>> # Compute pressure anistropy
    >>> p_anis = pyrf.pres_anis(p_xyz_i, b_xyz)

    """

    assert p_xyz is not None and isinstance(p_xyz, xr.DataArray)
    assert b_xyz is not None and isinstance(b_xyz, xr.DataArray)

    b_xyz = resample(b_xyz, p_xyz)

    # rotate pressure tensor to field aligned coordinates
    p_xyzfac = rotate_tensor(p_xyz, "fac", b_xyz, "pp")

    # Get parallel and perpendicular pressure
    p_para = p_xyzfac[:, 0, 0]
    p_perp = (p_xyzfac[:, 1, 1] + p_xyzfac[:, 2, 2]) / 2

    # Load permittivity
    mu0 = constants.mu0.value

    # Compute pressure anistropy
    p_anis = 1e9 * mu0 * (p_para - p_perp) / np.linalg.norm(b_xyz) ** 2

    return p_anis



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
ebsp.py

@author : Louis RICHARD
"""

import bisect
import warnings
import multiprocessing as mp
import numpy as np
import xarray as xr
import pyfftw
import sfs

from astropy.time import Time
from tqdm import tqdm

from .ts_time import ts_time
from .ts_vec_xyz import ts_vec_xyz
from .resample import resample
from .iso2unix import iso2unix
from .start import start
from .end import end
from .calc_fs import calc_fs
from .convert_fac import convert_fac

# TODO parallelized average_data


def average_data(data=None, x=None, y=None, av_window=None):
    # average data with time x to time y using window

    dtx, dty = [np.median(np.diff(x)), np.median(np.diff(y))]

    if av_window is None:
        av_window = dty

    dt2 = av_window / 2

    n_data_out = len(y)

    # Pad data with NaNs from each side
    n_point_to_add = int(np.ceil(dt2 / dtx))
    pad_nan = np.zeros((n_point_to_add, data.shape[1])) * np.nan
    data = np.vstack([pad_nan, data, pad_nan])
    pad_time = dtx * np.arange(n_point_to_add)
    x = np.hstack([x[0] - np.flip(pad_time), x, x[-1] + pad_time])

    out = np.zeros((n_data_out, data.shape[1]), dtype="complex128")

    for i, iy in enumerate(y):
        il = bisect.bisect_left(x, iy - dt2)
        ir = bisect.bisect_left(x, iy + dt2)

        out[i, :] = np.nanmean(data[il:ir, :], axis=0)

    return out


# noinspection PyUnboundLocalVariable
def ebsp(e=None, db=None, full_b=None, b0=None, xyz=None, freq_int=None, **kwargs):
    """Calculates wavelet spectra of E&B and Poynting flux using wavelets (Morlet wavelet). Also
    computes polarization parameters of B using SVD [7]_. SVD is performed on spectral matrices
    computed from the time series of B using wavelets and then averaged over a number of wave
    periods.

    Parameters
    ----------
    e : xarray.DataArray
        Time series of the wave electric field.

    db : xarray.DataArray
        Time series of the wave magnetic field.

    full_b : xarray.DataArray
        Time series of the high resolution background magnetic field used for E.B=0.

    b0 : xarray.DataArray
        Time series of the background magnetic field used for field aligned coordinates.

    xyz : xarray.DataArray
        Time series of the position time series of spacecraft used for field aligned coordinates.

    freq_int : str or list or numpy.ndarray
        Frequency interval :
            * "pc12" : [0.1, 5.0]
            * "pc35" : [2e-3, 0.1]
            * [fmin, fmax] : arbitrary interval [fmin,fmax]

    **kwargs : dict
        Hash table of keyword arguments with :
            * polarization : bool
                Computes polarization parameters. Default False.

            * noresamp : bool
                No resampling, E and db are given at the same time line Default False.

            * fac : bool
                Uses FAC coordinate system (defined by b0 and optionally xyz), otherwise no
                coordinate system transformation is performed. Default False.

            * de_dot_b0 : bool
                Computes dEz from db dot B = 0, uses full_b. Default False.

            * full_b_db : bool
                db contains DC field. Default False.

            * nav : int
                Number of wave periods to average Default 8.

            * fac_matrix : numpy.ndarray
                Specify rotation matrix to FAC system Default None.

            * m_width_coeff : int or float
                Specify coefficient to multiple Morlet wavelet width by. Default 1.

    Returns
    -------
    res : xarray.Dataset
        Dataset with :
            * t : xarray.DataArray
                Time.

            * f : xarray.DataArray
                Frequencies.

            * bb_xxyyzzss : xarray.DataArray
                db power spectrum with :
                    * [...,0] : x
                    * [...,1] : y
                    * [...,2] : z
                    * [...,3] : sum

            * ee_xxyyzzss : xarray.DataArray
                E power spectrum with :
                    * [...,0] : x
                    * [...,1] : y
                    * [...,2] : z
                    * [...,3] : sum

            * ee_ss : xarray.DataArray
                E power spectrum (xx+yy spacecraft coordinates, e.g. ISR2).

            * pf_xyz : xarray.DataArray
                Poynting flux (xyz).

            * pf_rtp : xarray.DataArray
                Poynting flux (r, theta, phi) [angles in degrees].

            * dop : xarray.DataArray
                3D degree of polarization.

            * dop2d : xarray.DataArray
                2D degree of polarization in the polarization plane.

            * planarity : xarray.DataArray
                Planarity of polarization.

            * ellipticity : xarray.DataArray
                Ellipticity of polarization ellipse.

            * k : xarray.DataArray
                k-vector (theta, phi FAC) [angles in degrees].


    See also
    --------
    pyrfu.plot.pl_ebsp : to fill.
    pyrfu.pyrf.convert_fac : Transforms to a field-aligned coordinate.

    Notes
    -----
    This software was developed as part of the MAARBLE (Monitoring, Analyzing and Assessing
    Radiation Belt Energization and Loss) collaborative research project which has received
    funding from the European Community's Seventh Framework Programme (FP7-SPACE-2011-1) under
    grant agreement n. 284520.

    References
    ----------
    .. [7] 	Santolík, O., Parrot. M., and  Lefeuvre. F. (2003) Singular value decomposition
            methods for wave propagation analysis,Radio Sci., 38(1), 1010,
            doi : https://doi.org/10.1029/2000RS002523 .

    Examples
    --------
    >>> from pyrfu import mms, pyrf
    >>> # Time interval
    >>> tint_brst = ["2015-10-30T05:15:42.000", "2015-10-30T05:15:54.000"]
    >>> # Spacecraft index
    >>> mms_id = 3
    >>> # Load spacecraft position
    >>> tint_long = pyrf.extend_tint(tint_brst, [-100, 100])
    >>> r_xyz = mms.get_data("R_gse", tint_long, mms_id)
    >>> # Load background magnetic field, electric field and magnetic field fluctuations
    >>> b_xyz = mms.get_data("B_gse_fgm_brst_l2", tint_brst, mms_id)
    >>> e_xyz = mms.get_data("E_gse_edp_brst_l2", tint_brst, mms_id)
    >>> b_scm = mms.get_data("B_gse_scm_brst_l2", tint_brst, mms_id)
    >>> # Polarization analysis
    >>> options = dict(polarization=True, fac=True)
    >>> polarization = pyrf.ebsp(e_xyz, b_scm, b_xyz, b_xyz, r_xyz, freq_int=[10, 4000], **options)

    """

    if not isinstance(db, xr.DataArray):
        raise TypeError("db must be a DataArray")

    if not isinstance(full_b, xr.DataArray):
        raise TypeError("full_b must be a DataArray")

    if not isinstance(b0, xr.DataArray):
        raise TypeError("b0 must be a DataArray")

    if not isinstance(xyz, xr.DataArray):
        raise TypeError("xyz must be a DataArray")

    # Number of threads
    n_threads = mp.cpu_count()

    # Compute magnetic field fluctuations sampling frequency
    fsb = calc_fs(db)

    # Check the input
    # Number of wave periods to average
    n_wave_period_to_average = 8

    # Below which we cannot apply E*B=0
    angle_b_elevation_max = 15

    # matrix for rotation to FAC
    fac_matrix = None

    m_width_coeff = 1

    want_polarization = False

    if e is None:
        want_ee = 0
    else:
        want_ee = 1

    res = dict(t=None, f=None, flagFac=0, bb_xxyyzzss=None, ee_xxyyzzss=None, ee_ss=None,
               pf_xyz=None, pf_rtp=None, dop=None, dop2d=None, planarity=None, ellipticity=None,
               k_tp=None, full_b=full_b, b0=b0, r=xyz)

    flag_no_resample, flag_want_fac, flag_de_dot_b0, flag_full_b_db = [False, False, False, False]

    if "polarization" in kwargs:
        want_polarization = True

    if "m_width_coeff" in kwargs:
        if not kwargs["m_width_coeff"] is None and isinstance(kwargs["m_width_coeff"], float):
            m_width_coeff = kwargs["m_width_coeff"]
        else:
            raise ValueError("parameter ''m_width_coeff'' without parameter value")

    if "no_resample" in kwargs:
        flag_no_resample = True

    if "fac" in kwargs:
        flag_want_fac = True

    if "de_dot_b0" in kwargs:
        flag_de_dot_b0 = True

    if "full_b_db" in kwargs:
        flag_full_b_db = True

    if "nav" in kwargs:
        if kwargs["nav"] is None or not isinstance(kwargs["nav"], int):
            raise TypeError("NAV requires  must be an integer")

        n_wave_period_to_average = kwargs["nav"]

    if "fac_matrix" in kwargs:
        if kwargs["fac_matrix"] is None or not isinstance(kwargs["fac_matrix"],
                            xr.DataArray) or "time" not in kwargs["fac_matrix"].coords:
            raise ValueError("fac_matrix requires a second argument struct(t,rotMatrix)")

        fac_matrix = kwargs["fac_matrix"]

    if flag_want_fac and fac_matrix is None:
        if b0 is None:
            raise ValueError("ebsp(): at least b0 should be given for option FAC")

        if xyz is None:
            print("convert_fac : assuming s/c position [1 0 0] for estimating FAC")
            xyz = [1, 0, 0]
            xyz = ts_vec_xyz(db.time.data, np.tile(xyz, (len(db), 1)))

        xyz = resample(xyz, db, fs=fsb)

    b0 = resample(b0, db, fs=fsb)

    if flag_full_b_db:
        full_b = db
        res["full_b"] = full_b
        db = db - b0

    if flag_de_dot_b0 and full_b is None:
        raise ValueError("full_b must be given for option de_dot_b0=0")

    pc12_range, pc35_range, other_range = [0, 0, 0]

    if isinstance(freq_int, str):
        if freq_int.lower() == "pc12":
            pc12_range = 1

            freq_int = [.1, 5]

            delta_t = 1

            tint = list(Time(np.round([start(db), end(db)]), format="unix").iso)

        elif freq_int.lower() == "pc35":
            pc35_range = 1

            freq_int = [.002, .1]

            delta_t = 60

            tint = list(
                Time([np.round(start(db) / 60), np.round(end(db) / 60)] * 60, format="unix").iso)
        else:
            raise ValueError("Invalid format of interval")

        out_sampling = 1 / delta_t

        nt = np.round((iso2unix(tint[1]) - iso2unix(tint[0])) / delta_t).astype(int)

        out_time = np.linspace(iso2unix(tint[0]), iso2unix(tint[1]), nt) + delta_t / 2
        out_time = out_time[:-1]
    else:
        if freq_int[1] >= freq_int[0]:
            other_range = True

            out_sampling = freq_int[1]/5

            delta_t = 1 / out_sampling

            nt = np.round((end(db) - start(db)) / delta_t).astype(int)

            out_time = np.linspace(start(db), end(db), nt) + delta_t / 2
            out_time = out_time[:-1]
        else:
            raise ValueError("FREQ_INT must be [f_min f_max], f_min<f_max")

    if want_ee:  # Check the sampling rate
        if e is None:
            raise ValueError("E cannot be empty for the chosen output parameters")

        fs_e, fs_b = [calc_fs(e), calc_fs(db)]

        if flag_no_resample:
            if fs_e != fs_b:
                raise IndexError("E and B must have the same sampling for no_resample")

            in_sampling = fs_e
        else:
            if fs_b > 1.5 * fs_e:
                e = resample(e, db, fs=fsb)

                b0 = resample(b0, db, fs=fsb)

                in_sampling = fs_b
                warnings.warn("Interpolating e to b", UserWarning)
            elif fs_e > 1.5 * fs_b:
                db = resample(db, e)
                b0 = resample(b0, e)

                in_sampling = fs_e
                warnings.warn("Interpolating b to e", UserWarning)
            elif fs_e == fs_b and len(e) == len(db):
                in_sampling = fs_e
            else:
                in_sampling = 2 * fs_e

                nt = (np.min([end(e), end(db)]) - np.max([start(e), start(db)])) / (1 / in_sampling)
                t = np.linspace(np.max([start(e), start(db)]), np.min([end(e), end(db)]), int(nt))

                t = ts_time(t)

                e, db, b0, full_b = [resample(field, t) for field in [e, db, b0, full_b]]

                warnings.warn("Interpolating b and e to 2x e sampling", UserWarning)

        print(f"Fs = {in_sampling:4.2f}, Fs_e = {fs_e:4.2f}, Fs_b = {fs_b:4.2f}")

    else:
        in_sampling = calc_fs(db)

        e = None

    if in_sampling / 2 < freq_int[1]:
        raise ValueError("F_MAX must be lower than the Nyquist frequency")

    if want_ee and e.shape[1] < 3 and not flag_de_dot_b0:
        raise ValueError("E must have all 3 components or flag ''de_dot_db=0'' must be given")

    if len(db) % 2:
        db = db[:-1, :]
        b0 = b0[:-1, :]

        if fac_matrix is None:
            xyz = xyz[:-1, :]
        else:
            fac_matrix["t"] = fac_matrix["t"][:-1, :]

            fac_matrix["rotMatrix"] = fac_matrix["rotMatrix"][:-1, :, :]

        if want_ee:
            e = e[:-1, :]

    in_time = db.time.data.view("i8")*1e-9

    b_x, b_y, b_z = [None, None, None]

    idx_b_par_spin_plane = None

    if flag_de_dot_b0:
        b_x = full_b[:, 0].data
        b_y = full_b[:, 1].data
        b_z = full_b[:, 2].data

        # Remove the last sample if the total number of samples is odd
        if len(full_b) % 2:
            b_x = b_x[:-1, :]
            b_y = b_y[:-1, :]
            b_z = b_z[:-1, :]

        angle_b_elevation = np.arctan(b_z / np.sqrt(b_x ** 2 + b_y ** 2)) * 180 / np.pi
        idx_b_par_spin_plane = np.abs(angle_b_elevation) < angle_b_elevation_max

    # If E has all three components, transform E and B waveforms to a magnetic field aligned
    # coordinate (FAC) and save eisr for computation of e_sum. Otherwise we compute Ez within the
    # main loop and do the transformation to FAC there.

    time_b0 = 0
    if flag_want_fac:
        res["flagFac"] = True

        time_b0 = b0.time.data.view("i8")*1e-9

        if want_ee:
            if not flag_de_dot_b0:
                eisr2 = e[:, :2]

                if e.shape[1] < 3:
                    raise TypeError("E must be a 3D vector to be rotated to FAC")

                if fac_matrix is None:
                    e = convert_fac(e, b0, xyz)
                else:
                    e = convert_fac(e, fac_matrix)

        if fac_matrix is None:
            db = convert_fac(db, b0, xyz)
        else:
            db = convert_fac(db, fac_matrix)

    # Find the frequencies for an FFT of all data and set important parameters
    nd2 = len(in_time)/2

    nyq = 0.5

    freq = in_sampling * np.arange(nd2) / nd2 * nyq

    w = np.hstack([0, freq, -np.flip(freq[:-1])])    # The frequencies corresponding to FFT

    morlet_width = 5.36 * m_width_coeff

    # to get proper overlap for Morlet
    freq_number = np.ceil((np.log10(freq_int[1]) - np.log10(freq_int[0])) * 12 * m_width_coeff)

    a_min, a_max = [np.log10(0.5 * in_sampling / freq_int[1]),
                    np.log10(0.5 * in_sampling / freq_int[0])]
    # a_min, a_max = [0.01, 2]

    a_number = freq_number
    # a_number = 400

    a = np.logspace(a_min, a_max, int(a_number))

    # Maximum frequency
    w0 = in_sampling / 2

    # Width of the Morlet wavelet
    sigma = morlet_width / w0
    # sigma = 5.36 / w0

    # Make the FFT of all data
    idx_nan_b = np.isnan(db.data)

    db.data[idx_nan_b] = 0

    swb = pyfftw.interfaces.numpy_fft.fft(db, axis=0, threads=n_threads)

    idx_nan_e, idx_nan_eisr2 = [None, None]

    sw_e, sw_eisr2 = [None, None]

    if want_ee:
        print("ebsp ... calculate E and B wavelet transform ... ")

        idx_nan_e = np.isnan(e.data)

        e.data[idx_nan_e] = 0

        sw_e = pyfftw.interfaces.numpy_fft.fft(e, axis=0, threads=n_threads)

        if flag_want_fac and not flag_de_dot_b0:
            idx_nan_eisr2 = np.isnan(eisr2.data)

            eisr2.data[idx_nan_eisr2] = 0

            sw_eisr2 = pyfftw.interfaces.numpy_fft.fft(eisr2, axis=0, threads=n_threads)
    else:
        print("ebsp ... calculate B wavelet transform ....")

    # Loop through all frequencies
    n_data, n_freq, n_data_out = [len(in_time), len(a), len(out_time)]

    #
    power_ex_plot, power_ey_plot, power_ez_plot = [np.zeros((n_data, n_freq), dtype="complex128")
                                                   for _ in range(3)]

    power_2e_plot, power_2e_isr2_plot = [np.zeros((n_data, n_freq), dtype="complex128") for _ in
                                         range(2)]

    power_bx_plot, power_by_plot, power_bz_plot = [np.zeros((n_data, n_freq), dtype="complex128")
                                                   for _ in range(3)]

    power_2b_plot = np.zeros((n_data, n_freq), dtype="complex128")

    s_plot_x, s_plot_y, s_plot_z = [np.zeros((n_data, n_freq)) for _ in range(3)]

    planarity, ellipticity = [np.zeros((n_data_out, n_freq)) for _ in range(2)]
    dop_3d, dop_2d = [np.zeros((n_data_out, n_freq), dtype="complex128") for _ in range(2)]

    theta_svd_fac, phi_svd_fac = [np.zeros((n_data_out, n_freq)) for _ in range(2)]

    # Get the correct frequencies for the wavelet transform
    frequency_vec = w0/a

    censure = np.floor(2 * a * out_sampling / in_sampling * n_wave_period_to_average)

    """
    pdb.set_trace()
    mpi_args = (frequency_vec,n_wave_period_to_average,out_sampling,sigma,a,w,w0,Swb,Swe,
    flag_want_fac,flag_de_dot_b0, Bx,By,Bz,b0,xyz,fac_matrix,in_time,out_time,idxNanB,idxNanE,
    idx_nan_eisr2,idx_b_par_spinplane,wantEE,wantPolarization,sw_eisr2)
    pool = mp.Pool(mp.cpu_count())
    pool.starmap(my_func, [(*mpi_args,ind_a) for ind_a in range(2)])
    """
    # begin for
    # inputs (frequency_vec,n_wave_period_to_average,out_sampling,sigma,w,w0,Swb,)

    for ind_a in tqdm(range(len(a))):
        # resample to 1 second sampling for Pc1-2 or 1 minute sampling for Pc3-5
        # average top frequencies to 1 second/1 minute
        # below will be an average over 8 wave periods. first find where one
        # sample is less than eight wave periods

        if frequency_vec[ind_a] / n_wave_period_to_average > out_sampling:
            av_window = 1 / out_sampling
        else:
            av_window = n_wave_period_to_average / frequency_vec[ind_a]

        # Get the wavelet transform by backward FFT
        w_exp_mat = np.exp(-sigma * sigma * ((a[ind_a] * w - w0) ** 2) / 2)
        w_exp_mat2 = np.tile(w_exp_mat, (2, 1)).T
        w_exp_mat = np.tile(w_exp_mat, (3, 1)).T

        wb = pyfftw.interfaces.numpy_fft.ifft(np.sqrt(1) * swb * w_exp_mat, axis=0,
                                              threads=n_threads)
        wb[idx_nan_b] = np.nan

        we, w_eisr2 = [None, None]

        if want_ee:
            if sw_e.shape[1] == 2:
                we = pyfftw.interfaces.numpy_fft.ifft(np.sqrt(1) * sw_e * w_exp_mat2, axis=0,
                                                      threads=n_threads)
            else:
                we = pyfftw.interfaces.numpy_fft.ifft(np.sqrt(1) * sw_e * w_exp_mat, axis=0,
                                                      threads=n_threads)

            we[idx_nan_e] = np.nan

            if flag_want_fac and not flag_de_dot_b0:
                w_eisr2 = pyfftw.interfaces.numpy_fft.ifft(np.sqrt(1) * sw_eisr2 * w_exp_mat2,
                                                           axis=0, threads=n_threads)
                w_eisr2[idx_nan_eisr2] = np.nan

        new_freq_mat = w0 / a[ind_a]
        # Power spectrum of E and Poynting flux

        if want_ee:
            # Power spectrum of E, power = (2*pi)*conj(W).*W./new_freq_mat
            if flag_want_fac and not flag_de_dot_b0:
                sum_power_eisr2 = np.sum(2 * np.pi * (w_eisr2 * np.conj(w_eisr2)) / new_freq_mat,
                                         axis=1)
            else:
                sum_power_eisr2 = np.sum(2 * np.pi * (we * np.conj(we)) / new_freq_mat, axis=1)

            power_2e_isr2_plot[:, ind_a] = sum_power_eisr2

            # Compute Ez from dE * B = 0
            if flag_de_dot_b0:
                we_re, we_im = [np.real(we), np.imag(we)]

                we_z = -(we_re[:, 0] * b_x + we_re[:, 1] * b_y) / b_z - 1j * (
                            we_im[:, 0] * b_x + we_im[:, 1] * b_y) / b_z
                we_z[idx_b_par_spin_plane] = np.nan

                if flag_want_fac:
                    if fac_matrix is None:
                        we = convert_fac(ts_vec_xyz(time_b0, np.hstack([we[:, :2], we_z])), b0, xyz)
                    else:
                        we = convert_fac(ts_vec_xyz(time_b0, np.hstack([we[:, :2], we_z])),
                                         fac_matrix)

                    we = we[:, 1:]
                else:
                    we = np.hstack([we[:, :2], we_z])

            power_e = 2 * np.pi * (we * np.conj(we)) / new_freq_mat
            power_e = np.vstack([power_e.T, np.sum(power_e, axis=1)]).T

            power_ex_plot[:, ind_a], power_ey_plot[:, ind_a], power_ez_plot[:, ind_a] = [
                power_e[:, i] for i in range(3)]

            power_2e_plot[:, ind_a] = power_e[:, 3]

            # Poynting flux calculations, assume E and b units mV/m and nT, get  S in uW/m^2
            # 4pi from wavelets, see A. Tjulins power estimates
            coeff_poynting = 10 / 4 / np.pi * (1/4) * (4 * np.pi)

            s = np.zeros((n_data, 3))

            we_x, we_y, we_z = [we[:, i] for i in range(3)]
            wb_x, wb_y, wb_z = [wb[:, i] for i in range(3)]

            s[:, 0] = coeff_poynting * np.real(
                we_y * np.conj(wb_z) + np.conj(we_y) * wb_z - we_z * np.conj(wb_y) - np.conj(
                    we_z) * wb_y) / new_freq_mat
            s[:, 1] = coeff_poynting * np.real(
                we_z * np.conj(wb_x) + np.conj(we_z) * wb_x - we_x * np.conj(wb_z) - np.conj(
                    we_x) * wb_z) / new_freq_mat
            s[:, 2] = coeff_poynting * np.real(
                we_x * np.conj(wb_y) + np.conj(we_x) * wb_y - we_y * np.conj(wb_x) - np.conj(
                    we_y) * wb_x) / new_freq_mat

            s_plot_x[:, ind_a], s_plot_y[:, ind_a], s_plot_z[:, ind_a] = [s[:, i] for i in range(3)]

        # Power spectrum of B
        power_b = 2 * np.pi * (wb * np.conj(wb)) / new_freq_mat
        power_b = np.vstack([power_b.T, np.sum(power_b, axis=1)]).T

        power_bx_plot[:, ind_a], power_by_plot[:, ind_a], power_bz_plot[:, ind_a] = [power_b[:, i]
                                                                                     for i in
                                                                                     range(3)]

        power_2b_plot[:, ind_a] = power_b[:, 3]

        # Polarization parameters
        if want_polarization:
            # Construct spectral matrix and average it
            s_mat = np.zeros((3, 3, n_data), dtype="complex128")

            for i in range(3):
                for j in range(3):
                    s_mat[i, j, :] = 2 * np.pi * (wb[:, i] * np.conj(wb[:, j])) / new_freq_mat

            s_mat = np.transpose(s_mat, [2, 0, 1])

            # Averaged s_mat
            s_mat_avg = np.zeros((n_data_out, 3, 3), dtype="complex128")

            for comp in range(3):
                s_mat_avg[..., comp] = average_data(s_mat[..., comp], in_time, out_time, av_window)

            # Remove data possibly influenced by edge effects
            censure_idx = np.hstack([np.arange(np.min([censure[ind_a], len(out_time)])),
                                     np.arange(np.max([0, len(out_time) - censure[ind_a] - 1]),
                                               len(out_time))])
            censure_idx = censure_idx.astype(int)

            s_mat_avg[censure_idx, ...] = np.nan

            # compute singular value decomposition
            # real matrix which is superposition of real part of spectral matrix over imaginary part
            a_mat, u_mat = [np.zeros((6, 3, n_data_out)) for _ in range(2)]
            w_mat, v_mat = [np.zeros((3, 3, n_data_out)) for _ in range(2)]

            # wSingularValues = zeros(3,n_data2);
            # R = zeros(3,3,n_data2); #spectral matrix in coordinate defined by V axes
            a_mat[:3, ...] = np.real(np.transpose(s_mat_avg, [1, 2, 0]))
            a_mat[3:6, ...] = -np.imag(np.transpose(s_mat_avg, [1, 2, 0]))

            for i in range(n_data_out):
                if np.isnan(a_mat[..., i]).any():
                    u_mat[..., i], w_mat[..., i], v_mat[..., i] = [np.nan, np.nan, np.nan]
                else:
                    u_mat[..., i], w_mat[..., i], v_mat[..., i] = np.linalg.svd(a_mat[..., i],
                                                                                full_matrices=False)

            # compute direction of propagation
            sign_kz = np.sign(v_mat[2, 2, :])
            v_mat[2, 2, :] = v_mat[2, 2, :] * sign_kz
            v_mat[1, 2, :] = v_mat[1, 2, :] * sign_kz
            v_mat[0, 2, :] = v_mat[0, 2, :] * sign_kz

            theta_svd_fac[:, ind_a] = np.abs(np.squeeze(
                np.arctan(np.sqrt(v_mat[0, 2, :] ** 2 + v_mat[1, 2, :] ** 2) / v_mat[2, 2, :])))

            phi_svd_fac[:, ind_a] = np.squeeze(np.arctan2(v_mat[1, 2, :], v_mat[0, 2, :]))

            # Calculate polarization parameters
            planarity_local = np.squeeze(1-np.sqrt(w_mat[2, 2, :] / w_mat[0, 0, :]))
            planarity_local[censure_idx] = np.nan

            planarity[:, ind_a] = planarity_local

            # ellipticity: ratio of axes of polarization ellipse axes*sign of polarization

            ellipticity_local = np.squeeze(w_mat[1, 1, :] / w_mat[0, 0, :]) * np.sign(
                np.imag(s_mat_avg[:, 0, 1]))
            ellipticity_local[censure_idx] = np.nan

            ellipticity[:, ind_a] = ellipticity_local

            # DOP = sqrt[(3/2.*trace(SM^2)./(trace(SM))^2 - 1/2)]; Samson, 1973, JGR
            dop = np.sqrt((3 / 2) * (np.trace(np.matmul(s_mat_avg, s_mat_avg), axis1=1, axis2=2)
                                     / np.trace(s_mat_avg, axis1=1, axis2=2) ** 2) - 1 / 2)

            dop[censure_idx] = np.nan
            dop_3d[:, ind_a] = dop

            # DOP in 2D = sqrt[2*trace(rA^2)/trace(rA)^2 - 1)]; Ulrich
            v_mat_new = np.transpose(v_mat, [2, 0, 1])

            s_mat_avg2dim = np.matmul(v_mat_new,
                                      np.matmul(s_mat_avg, np.transpose(v_mat_new, [0, 2, 1])))
            s_mat_avg2dim = s_mat_avg2dim[:, :2, :2]
            s_mat_avg = s_mat_avg2dim

            dop2dim = np.sqrt(2 * (
                        np.trace(np.matmul(s_mat_avg, s_mat_avg), axis1=1, axis2=2) / np.trace(
                s_mat_avg, axis1=1, axis2=2) ** 2) - 1)
            dop2dim[censure_idx] = np.nan
            dop_2d[:, ind_a] = dop

    # set data gaps to NaN and remove edge effects
    censure = np.floor(2*a)

    for ind_a in range(len(a)):
        censure_idx = np.hstack([np.arange(np.min([censure[ind_a], len(in_time)])), np.arange(
            np.max([1, len(in_time) - censure[ind_a]]), len(in_time))])

        censure_idx = censure_idx.astype(int)

        power_bx_plot[censure_idx, ind_a] = np.nan
        power_by_plot[censure_idx, ind_a] = np.nan
        power_bz_plot[censure_idx, ind_a] = np.nan
        power_2b_plot[censure_idx, ind_a] = np.nan

        if want_ee:
            power_ex_plot[censure_idx, ind_a] = np.nan
            power_ey_plot[censure_idx, ind_a] = np.nan
            power_ez_plot[censure_idx, ind_a] = np.nan
            power_2e_plot[censure_idx, ind_a] = np.nan

            power_2e_isr2_plot[censure_idx, ind_a] = np.nan

            s_plot_x[censure_idx, ind_a] = np.nan
            s_plot_y[censure_idx, ind_a] = np.nan
            s_plot_z[censure_idx, ind_a] = np.nan

    # remove edge effects from data gaps
    idx_nan_e = np.sum(idx_nan_e, axis=1) > 0
    idx_nan_b = np.sum(idx_nan_b, axis=1) > 0
    idx_nan_eisr2 = np.sum(idx_nan_eisr2, axis=1) > 0

    n_data2 = len(power_2b_plot)
    if pc12_range or other_range:
        censure3 = np.floor(1.8 * a)
    elif pc35_range:
        censure3 = np.floor(.4 * a)
    else:
        raise ValueError("Invalid range")

    for i in range(len(idx_nan_b) - 1):
        if idx_nan_b[i] < idx_nan_b[i + 1]:
            for j in range(len(a)):
                censure_index_front = np.arange(np.max([i - censure3[j], 0]), i)

                power_bx_plot[censure_index_front, j] = np.nan
                power_by_plot[censure_index_front, j] = np.nan
                power_bz_plot[censure_index_front, j] = np.nan
                power_2b_plot[censure_index_front, j] = np.nan

                s_plot_x[censure_index_front, j] = np.nan
                s_plot_y[censure_index_front, j] = np.nan
                s_plot_z[censure_index_front, j] = np.nan

        if idx_nan_b[i] > idx_nan_b[i + 1]:
            for j in range(len(a)):
                censure_index_back = np.arange(i, np.min([i + censure3[j], n_data2]))

                power_bx_plot[censure_index_back, j] = np.nan
                power_by_plot[censure_index_back, j] = np.nan
                power_bz_plot[censure_index_back, j] = np.nan
                power_2b_plot[censure_index_back, j] = np.nan

                s_plot_x[censure_index_back, j] = np.nan
                s_plot_y[censure_index_back, j] = np.nan
                s_plot_z[censure_index_back, j] = np.nan

    n_data3 = len(power_2e_plot)

    for i in range(len(idx_nan_e) - 1):
        if idx_nan_e[i] < idx_nan_e[i + 1]:
            for j in range(len(a)):
                censure_index_front = np.arange(np.max([i - censure3[j], 1]), i)

                power_ex_plot[censure_index_front, j] = np.nan
                power_ey_plot[censure_index_front, j] = np.nan
                power_ez_plot[censure_index_front, j] = np.nan
                power_2e_plot[censure_index_front, j] = np.nan

                power_2e_isr2_plot[censure_index_front, j] = np.nan

                s_plot_x[censure_index_front, j] = np.nan
                s_plot_y[censure_index_front, j] = np.nan
                s_plot_z[censure_index_front, j] = np.nan

        elif idx_nan_e[i] > idx_nan_e[i + 1]:
            for j in range(len(a)):
                censure_index_back = np.arange(i, np.min([i+censure3[j], n_data3]))

                power_ex_plot[censure_index_back, j] = np.nan
                power_ey_plot[censure_index_back, j] = np.nan
                power_ez_plot[censure_index_back, j] = np.nan
                power_2e_plot[censure_index_back, j] = np.nan

                power_2e_isr2_plot[censure_index_back, j] = np.nan

                s_plot_x[censure_index_back, j] = np.nan
                s_plot_y[censure_index_back, j] = np.nan
                s_plot_z[censure_index_back, j] = np.nan

        else:
            continue

    n_data4 = len(power_2e_isr2_plot)

    for i in range(len(idx_nan_eisr2) - 1):
        if idx_nan_eisr2[i] < idx_nan_eisr2[i + 1]:
            for j in range(len(a)):
                censure_index_front = np.arange(np.max([i - censure3[j], 0]), i)

                power_2e_isr2_plot[censure_index_front, j] = np.nan

        elif idx_nan_eisr2[i] > idx_nan_eisr2[i + 1]:
            for j in range(len(a)):
                censure_index_back = np.arange(i, np.min([i + censure3[j], n_data4]))

                power_2e_isr2_plot[censure_index_back, j] = np.nan

    power_bx_plot = average_data(power_bx_plot, in_time, out_time)
    power_by_plot = average_data(power_by_plot, in_time, out_time)
    power_bz_plot = average_data(power_bz_plot, in_time, out_time)
    power_2b_plot = average_data(power_2b_plot, in_time, out_time)

    bb_xxyyzzss = np.tile(power_bx_plot, (4, 1, 1))
    bb_xxyyzzss = np.transpose(bb_xxyyzzss, [1, 2, 0])
    bb_xxyyzzss[:, :, 1] = power_by_plot
    bb_xxyyzzss[:, :, 2] = power_bz_plot
    bb_xxyyzzss[:, :, 3] = power_2b_plot
    bb_xxyyzzss = np.real(bb_xxyyzzss)

    # Output
    res["t"] = Time(out_time, format="unix").datetime64
    res["f"] = frequency_vec
    res["bb_xxyyzzss"] = xr.DataArray(bb_xxyyzzss,
                                      coords=[res["t"], res["f"], ["xx", "yy", "zz", "ss"]],
                                      dims=["time", "frequency", "comp"])

    if want_ee:
        power_ex_plot = average_data(power_ex_plot, in_time, out_time)
        power_ey_plot = average_data(power_ey_plot, in_time, out_time)
        power_ez_plot = average_data(power_ez_plot, in_time, out_time)
        power_2e_plot = average_data(power_2e_plot, in_time, out_time)

        power_2e_isr2_plot = average_data(power_2e_isr2_plot, in_time, out_time)
        power_2e_isr2_plot = np.real(power_2e_isr2_plot)

        s_plot_x = np.real(average_data(s_plot_x, in_time, out_time))
        s_plot_y = np.real(average_data(s_plot_y, in_time, out_time))
        s_plot_z = np.real(average_data(s_plot_z, in_time, out_time))
        s_azimuth, s_elevation, s_r = sfs.util.cart2sph(s_plot_x, s_plot_y, s_plot_z)

        ee_xxyyzzss = np.tile(power_ex_plot, (4, 1, 1))
        ee_xxyyzzss = np.transpose(ee_xxyyzzss, [1, 2, 0])
        ee_xxyyzzss[:, :, 1] = power_ey_plot
        ee_xxyyzzss[:, :, 2] = power_ez_plot
        ee_xxyyzzss[:, :, 3] = power_2e_plot
        ee_xxyyzzss = np.real(ee_xxyyzzss)

        poynting_xyz = np.tile(s_plot_x, (3, 1, 1))
        poynting_xyz = np.transpose(poynting_xyz, [1, 2, 0])
        poynting_xyz[:, :, 1] = s_plot_y
        poynting_xyz[:, :, 2] = s_plot_z
        poynting_xyz = poynting_xyz.astype(float)

        poynting_r_th_ph = np.tile(s_r, (3, 1, 1))
        poynting_r_th_ph = np.transpose(poynting_r_th_ph, [1, 2, 0])
        poynting_r_th_ph[..., 1] = np.pi / 2 - s_elevation
        poynting_r_th_ph[..., 2] = s_azimuth
        poynting_r_th_ph[..., 1:] = poynting_r_th_ph[..., 1:] * 180 / np.pi
        poynting_r_th_ph = poynting_r_th_ph.astype(float)

        # Output
        res["ee_ss"] = power_2e_isr2_plot.astype(float)

        res["ee_xxyyzzss"] = xr.DataArray(ee_xxyyzzss,
                                          coords=[res["t"], res["f"], ["xx", "yy", "zz", "ss"]],
                                          dims=["time", "frequency", "comp"])

        res["pf_xyz"] = xr.DataArray(poynting_xyz, coords=[res["t"], res["f"], ["x", "y", "z"]],
                                     dims=["time", "frequency", "comp"])

        res["pf_rtp"] = xr.DataArray(poynting_r_th_ph,
                                     coords=[res["t"], res["f"], ["rho", "theta", "phi"]],
                                     dims=["time", "frequency", "comp"])

    if want_polarization:
        # Define parameters for which we cannot compute the wave vector
        with warnings.catch_warnings():
            warnings.simplefilter(action="ignore", category=RuntimeWarning)
            ind_low_planarity = planarity < 0.5
            ind_low_ellipticity = np.abs(ellipticity) < .2

        theta_svd_fac[ind_low_planarity] = np.nan
        phi_svd_fac[ind_low_planarity] = np.nan

        theta_svd_fac[ind_low_ellipticity] = np.nan
        phi_svd_fac[ind_low_ellipticity] = np.nan

        k_th_ph_svd_fac = np.zeros((theta_svd_fac.shape[0], theta_svd_fac.shape[1], 2))
        k_th_ph_svd_fac[..., 0] = theta_svd_fac
        k_th_ph_svd_fac[..., 1] = phi_svd_fac

        # Output
        res["dop"] = xr.DataArray(np.real(dop_3d), coords=[res["t"], res["f"]],
                                  dims=["time", "frequency"])

        res["dop2d"] = xr.DataArray(np.real(dop_2d), coords=[res["t"], res["f"]],
                                    dims=["time", "frequency"])

        res["planarity"] = xr.DataArray(planarity, coords=[res["t"], res["f"]],
                                        dims=["time", "frequency"])

        res["ellipticity"] = xr.DataArray(ellipticity, coords=[res["t"], res["f"]],
                                          dims=["time", "frequency"])

        res["k_tp"] = xr.DataArray(k_th_ph_svd_fac, coords=[res["t"], res["f"], ["theta", "phi"]],
                                   dims=["time", "frequency", "comp"])

    return res



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
make_model_vdf.py

@author : Louis RICHARD
"""

import numpy as np
import xarray as xr

from astropy import constants

from ..pyrf import resample, dec_par_perp, norm
from . import rotate_tensor


def make_model_vdf(vdf=None, b_xyz=None, sc_pot=None, n=None, v_xyz=None, t_xyz=None):
    """
    Make a general bi-Maxwellian distribution function based on particle moment data in the same
    format as PDist.

    Parameters
    ----------
    vdf : xarray.Dataset
        Particle distribution (skymap).

    b_xyz : xarray.DataArray
        Time series of the background magnetic field.

    sc_pot : xarray.DataArray
        Time series of the spacecraft potential.

    n : xarray.DataArray
        Time series of the number density.

    v_xyz : xarray.DataArray
        Time series of the bulk velocity.

    t_xyz : xarray.DataArray
        Time series of the temperature tensor.

    Returns
    -------
    model_vdf : xarray.Dataset
        Distribution function in the same format as vdf.

    See also
    --------
    pyrfu.mms.calculate_epsilon : Calculates epsilon parameter using model distribution.

    Examples
    --------
    >>> from pyrfu import mms
    >>> model_vdf = mms.make_model_vdf(vdf_e, b_xyz, sc_pot, n_e, v_xyz_e, t_xyz_e)

    """

    # Check that PDist and moments have the same times
    """
    if abs(median(diff(PDist.time-n.time))) > 0
        modelPDist = NaN;
        irf.log('critical','PDist and moments have different times.')
        return;
    end
    """

    assert vdf is not None and isinstance(vdf, xr.Dataset)
    assert b_xyz is not None and isinstance(b_xyz, xr.DataArray)
    assert sc_pot is not None and isinstance(sc_pot, xr.DataArray)
    assert n is not None and isinstance(n, xr.DataArray)
    assert v_xyz is not None and isinstance(v_xyz, xr.DataArray)
    assert t_xyz is not None and isinstance(t_xyz, xr.DataArray)

    # Convert to SI units
    vdf /= 1e18

    # Resample b_xyz and sc_pot to particle data resolution
    b_xyz, sc_pot = [resample(b_xyz, n), resample(sc_pot, n)]

    # Define directions based on b_xyz and v_xyz, calculate relevant temperatures
    t_xyzfac = rotate_tensor(t_xyz, "fac", b_xyz, "pp")  # N.B makes final distribution gyrotropic
    t_para, t_ratio = [t_xyzfac[:, 0, 0], t_xyzfac[:, 0, 0] / t_xyzfac[:, 1, 1]]

    v_para, v_perp, alpha = dec_par_perp(v_xyz, b_xyz)

    v_perp_mag, b_xyz_mag = [norm(v_perp), norm(b_xyz)]
    v_perp_dir, b_xyz_dir = [v_perp / v_perp_mag, b_xyz / b_xyz_mag]

    # Define constants
    qe = constants.e.value

    # Check whether particles are electrons or ions
    if vdf.attrs["species"].lower() == "e":
        p_mass = constants.m_e.value
        print("notice : Particles are electrons")
    elif vdf.attrs["species"].lower() == "i":
        p_mass = constants.m_p.value
        sc_pot.data = -sc_pot.data
        print("notice : Particles are ions")
    else:
        raise ValueError("Invalid specie")

    # Convert moments to SI units
    vth_para = np.sqrt(2 * t_para * qe / p_mass)

    v_perp_mag.data *= 1e3
    v_para.data *= 1e3
    n.data *= 1e6

    # Defines dimensions of array below
    n_ti = len(vdf.time)
    n_ph = len(vdf.phi[0, :])
    n_th = len(vdf.theta)
    n_en = len(vdf.attrs["energy0"])

    # Get energy array
    energy = vdf.energy

    # Define Cartesian coordinates
    x, y, z = [np.zeros((n_ti, n_ph, n_th)) for _ in range(3)]

    r = np.zeros((n_ti, n_en))

    for ii in range(n_ti):
        x[ii, ...] = np.outer(
            -np.cos(vdf.phi.data[ii, :] * np.pi / 180) * np.sin(vdf.theta.data * np.pi / 180))
        y[ii, ...] = np.outer(
            -np.sin(vdf.phi.data[ii, :] * np.pi / 180) * np.sin(vdf.theta.data * np.pi / 180))
        z[ii, ...] = np.outer(-np.ones(n_ph) * np.cos(vdf.theta.data * np.pi / 180))
        r[ii, ...] = np.real(np.sqrt(2 * (energy[ii, :] - sc_pot.data[ii]) * qe / p_mass))

    r[r == 0] = 0.0

    # Define rotation vectors based on B and Ve directions
    r_x, r_y, r_z = v_perp_dir.data, np.cross(b_xyz_dir.data, v_perp_dir.data), b_xyz_dir.data

    # Rotated coordinate system for computing bi-Maxwellian distribution
    x_p, y_p, z_p = [np.zeros((n_ti, n_ph, n_th)) for _ in range(3)]

    for ii in range(n_ti):
        x_p[ii, ...] = x[ii, ...] * r_x[ii, 0] + y[ii, ...] * r_x[ii, 1] + z[ii, ...] * r_x[ii, 2]
        y_p[ii, ...] = x[ii, ...] * r_y[ii, 0] + y[ii, ...] * r_y[ii, 1] + z[ii, ...] * r_y[ii, 2]
        z_p[ii, ...] = x[ii, ...] * r_z[ii, 0] + y[ii, ...] * r_z[ii, 1] + z[ii, ...] * r_z[ii, 2]

    # Make 4D position matrix
    x_p = np.tile(x_p, [n_en, 1, 1, 1])
    x_p = np.transpose(x_p, [1, 0, 2, 3])

    y_p = np.tile(x_p, [n_en, 1, 1, 1])
    y_p = np.transpose(y_p, [1, 0, 2, 3])

    z_p = np.tile(x_p, [n_en, 1, 1, 1])
    z_p = np.transpose(z_p, [1, 0, 2, 3])

    r_mat = np.tile(r, [n_ph, n_th, 1, 1])
    r_mat = np.transpose(r_mat, [2, 3, 0, 1])

    # Can use 4D matrices. Too much memory is used for my computer; too slow for large files.
    # Make 4D matrices required for distribution calculation
    # nmat = repmat(n.data,1,lengthenergy,lengthphi,lengththeta);
    # Tratmat = repmat(Trat.data,1,lengthenergy,lengthphi,lengththeta);
    # Vpmagmat = repmat(Vpmag.data,1,lengthenergy,lengthphi,lengththeta);
    # Vparmat = repmat(Vpar.data,1,lengthenergy,lengthphi,lengththeta);
    # vthparmat = repmat(vthpar.data,1,lengthenergy,lengthphi,lengththeta);

    # Calculate bi-Maxwellian distribution function
    # bimaxdist = nmat * Tratmat./(sqrt(pi^3)*vthparmat.^3);
    # bimaxdist = bimaxdist.*exp(-(xp.*rmat-Vpmagmat).^2./(vthparmat.^2).*Tratmat);
    # bimaxdist = bimaxdist.*exp(-(yp.*rmat).^2./(vthparmat.^2).*Tratmat);
    # bimaxdist = bimaxdist.*exp(-(zp.*rmat-Vparmat).^2./(vthparmat.^2));

    # Construct bi-Maxwellian distribution function
    bi_max_dist = np.zeros(r_mat.shape)

    for ii in range(n_ti):
        coeff = n.data[ii] * t_ratio.data[ii] / (np.sqrt(np.pi ** 3) * vth_para.data[ii] ** 3)

        bi_max_temp = coeff * np.exp(
            -(x_p[ii, ...] * r_mat[ii, ...] - v_perp_mag.data[ii]) ** 2 / (vth_para.data[ii] ** 2) *
            t_ratio.data[ii])
        bi_max_temp = bi_max_temp * np.exp(
            -(y_p[ii, ...] * r_mat[ii, ...]) ** 2 / (vth_para.data[ii] ** 2) * t_ratio.data[ii])
        bi_max_temp = bi_max_temp * np.exp(
            -(z_p[ii, ...] * r_mat[ii, ...] - v_para.data[ii]) ** 2 / (vth_para.data[ii] ** 2))

        bi_max_dist[ii, ...] = bi_max_temp

    # Make modelPDist file for output
    model_vdf = vdf
    model_vdf.data = bi_max_dist
    model_vdf *= 1e18

    return model_vdf




#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
mms_config.py

@author : Louis RICHARD
"""

CONFIG = {"local_data_dir": "/Users/louisr/Documents/PhD/solo",
          'mirror_data_dir': None,  # e.g., '/Volumes/data_network/data/mms'
          'debug_mode': False,
          'download_only': False,
          'no_download': False}


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
calc_dt.py

@author : Louis RICHARD
"""

import numpy as np
import xarray as xr


def calc_dt(inp=None):
    """Computes time step of the input time series.

    Parameters
    ----------
    inp : xarray.DataArray
        Time series of the input variable.

    Returns
    -------
    out : float
        Time step in seconds.

    """

    if inp is None:
        raise ValueError("calc_dt requires at least one argument")

    if not isinstance(inp, xr.DataArray):
        raise TypeError("Input must be a DataArray")

    out = np.median(np.diff(inp.time.data)).astype(float) * 1e-9

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
vdf_omni.py

@author : Louis RICHARD
"""

import numpy as np
import xarray as xr


def vdf_omni(vdf):
    """Computes omnidirectional distribution, conserving unit.

    Parameters
    ----------
    vdf : xarray.Dataset
        Time series of the 3D velocity distribution with :
            * time : Time samples.
            * data : 3D velocity distribution.
            * energy : Energy levels.
            * phi : Azimuthal angles.
            * theta : Elevation angle.

    Returns
    -------
    out : xarray.Dataset
        Time series of the omnidirectional velocity distribution function with :
            * time : Time samples.
            * data : 3D velocity distribution.
            * energy : Energy levels.

    """

    assert vdf is not None and isinstance(vdf, xr.Dataset)

    time = vdf.time.data

    energy = vdf.energy.data
    thetas = vdf.theta.data
    dangle = np.pi / 16
    np_phi = 32

    z2 = np.ones((np_phi, 1)) * np.sin(thetas * np.pi / 180)
    solid_angles = dangle * dangle * z2
    all_solid_angles = np.tile(solid_angles, (len(time), energy.shape[1], 1, 1))

    dist = vdf.data.data * all_solid_angles
    omni = np.squeeze(np.nanmean(np.nanmean(dist, axis=3), axis=2))
    omni /= np.mean(np.mean(solid_angles))

    energy = np.mean(energy[:2, :], axis=0)

    out = xr.DataArray(omni, coords=[time, energy], dims=["time", "energy"], attrs=vdf.attrs)

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
dist_append.py

@author : Louis RICHARD
"""

import numpy as np
import xarray as xr
from .ts_skymap import ts_skymap


def dist_append(inp0=None, inp1=None):
    """Concatenate two distribution skymaps along the time axis.

    Parameters
    ----------
    inp0 : xarray.DataArray
        3D skymap distribution at early times.

    inp1 : xarray.DataArray
        3D skymap distribution at late times.

    Returns
    -------
    out : xarray.DataArray
        3D skymap of the concatenated 3D skymaps.

    Notes
    -----
    The time series have to be in the correct time order.

    """

    assert inp0 is not None and isinstance(inp0, xr.Dataset)
    assert inp1 is not None and isinstance(inp1, xr.Dataset)

    # time
    time = np.hstack([inp0.time.data, inp1.time.data])

    # Azimuthal angle
    if inp0.phi.ndim == 2:
        phi = np.vstack([inp0.phi.data, inp1.phi.data])
    else:
        phi = inp0.phi.data

    # Elevation angle
    theta = inp0.theta.data

    # distribution
    data = np.vstack([inp0.data, inp1.data])

    if "delta_energy_plus" in inp0.attrs:
        delta_energy_plus = np.vstack(
            [inp0.attrs["delta_energy_plus"].data, inp1.attrs["delta_energy_plus"].data])
    else:
        delta_energy_plus = None

    if "delta_energy_minus" in inp0.attrs:
        delta_energy_minus = np.vstack(
            [inp0.attrs["delta_energy_minus"].data, inp1.attrs["delta_energy_minus"].data])
    else:
        delta_energy_minus = None

    # Energy
    if inp0.attrs["tmmode"] == "brst":
        step_table = np.hstack([inp0.attrs["esteptable"], inp1.attrs["esteptable"]])

        out = ts_skymap(time, data, None, phi, theta, energy0=inp0.energy0, energy1=inp0.energy1,
                        esteptable=step_table)
    else:
        energy = np.vstack([inp0.energy.data, inp1.energy.data])

        out = ts_skymap(time, data, energy, phi, theta)

    # attributes
    attrs = inp0.attrs

    if "esteptable" in attrs.keys():
        attrs.pop("esteptable")

    if "delta_energy_minus" in attrs.keys():
        attrs["delta_energy_minus"] = delta_energy_minus

    if "delta_energy_plus" in attrs.keys():
        attrs["delta_energy_plus"] = delta_energy_plus

    for k in attrs:
        out.attrs[k] = attrs[k]

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
split_vs.py

@author : Louis RICHARD
"""

import warnings


def split_vs(var_str=""):
    """Parse the variable keys.

    Parameters
    ----------
    var_str : str
        Input key of variable

    Returns
    -------
    out : dict
        Hash table containing :
            * param : Variable key.
            * to : Tensor order.
            * cs : Coordinate system.
            * inst : Instrument.
            * tmmode : Time mode.
            * lev" : Level of data.

    """

    assert var_str and isinstance(var_str, str)

    tk = var_str.split("_")
    n_tk = len(tk)

    if n_tk < 3 or n_tk > 5:
        raise ValueError("invalid STRING format")

    all_params_scalars = ["ni", "nbgi", "pbgi", "partni", "ne", "pbge", "nbge", "partne", "tsi",
                          "tperpi",
                          "tparai", "parttperpi", "parttparai", "tse", "tperpe", "tparae",
                          "parttperpe", "parttparae",
                          "pde", "pdi", "pderre", "pderri", "v", "v6", "defi", "defbgi", "defe",
                          "defbge",
                          "energyi", "bnergye", "epar", "sdev12", "sdev34", "flux-amb-pm2",
                          "padlowene", "padmidene",
                          "padhighene", "bpsd", "epsd"]

    all_params_vectors = ["r", "sti", "vi", "errvi", "partvi", "ste", "ve", "errve", "partve", "qi",
                          "errqi", "qe",
                          "errqe", "b", "e", "e2d", "es12", "es34"]

    all_params_tensors = ["pi", "partpi", "pe", "partpe", "ti", "partti", "te", "partte"]

    hpca_params_scalars = ["nhplus", "nheplus", "nheplusplus", "noplus", "tshplus", "tsheplus",
                           "tsheplusplus", "tsoplus",
                           "phase", "adcoff"]

    hpca_params_tensors = ["vhplus", "vheplus", "vheplusplus", "voplus", "phplus", "pheplus",
                           "pheplusplus", "poplus",
                           "thplus", "theplus", "theplusplus", "toplus"]

    instruments = ["mec", "fpi", "edp", "edi", "hpca", "fgm", "dfg", "afg", "scm", "fsm", "dsp"]

    coordinate_systems = ["gse", "gsm", "dsl", "dbcs", "dmpa", "ssc", "bcs", "par"]

    data_lvls = ["ql", "sitl", "l1b", "l2a", "l2pre", "l2", "l3"]

    param = tk[0]

    if param.lower() in all_params_scalars:
        tensor_order = 0
    elif param.lower() in all_params_vectors:
        tensor_order = 1
    elif param.lower() in all_params_tensors:
        tensor_order = 2
    elif param.lower() in hpca_params_scalars:
        tensor_order = 0
    elif param.lower() in hpca_params_tensors:
        tensor_order = 1
    else:
        raise ValueError(f"invalid PARAM : {param}")

    coordinate_system = []
    idx = 0

    if tensor_order > 0:
        coordinate_system = tk[idx+1]

        idx += 1

        if coordinate_system not in coordinate_systems:
            raise ValueError("invalid COORDINATE_SYS")

    instrument = tk[idx+1]
    idx += 1

    if instrument not in instruments:
        raise ValueError("invalid INSTRUMENT")

    tmmode = tk[idx+1]
    idx += 1

    if tmmode not in ["brst", "fast", "slow", "srvy"]:
        tmmode = "fast"
        idx -= 1
        warnings.warn("assuming TM_MODE = FAST", UserWarning)

    if len(tk) == idx + 1:
        data_lvl = "l2"  # default
    else:
        data_lvl = tk[idx+1]

        if data_lvl not in data_lvls:
            raise ValueError("invalid DATA_LEVEL level")

    res = {"param": param, "to": tensor_order, "cs": coordinate_system, "inst": instrument,
           "tmmode": tmmode, "lev": data_lvl}

    return res



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
fft_bandpass.py

@author : Louis RICHARD
"""

import numpy as np
import xarray as xr

from ..pyrf import calc_fs, ts_scalar, ts_vec_xyz


def fft_bandpass(inp=None, f_min=None, f_max=None):
    """Perform simple bandpass using FFT - returns fields between with ``f_min`` < f < ``f_max``.

    Parameters
    ----------
    inp : xarray.DataArray
        Time series to be bandpassed filtered.

    f_min : float or int
        Minimum frequency of filter, f < ``f_min`` are removed.

    f_max : float or int
        Maximum frequency of filter, f > ``f_max`` are removed.

    Returns
    -------
    out : xarray.DataArray
        Time series of the bandpassed filtered data.

    Notes
    -----
    Can be some spurius effects near boundary. Can take longer interval then use tlim to remove.

    Examples
    --------
    >>> from pyrfu import mms
    >>> b_xyz_bp = mms.fft_bandpass(e_xyz, f_min, f_max)

    """

    assert inp is not None and isinstance(inp, xr.DataArray)
    assert f_min is not None and isinstance(f_min, (float, int))
    assert f_max is not None and isinstance(f_max, (float, int))

    inp_time, inp_data = [inp.time.data, inp.data]

    # Make sure number of elements is an even number, if odd remove last element to make an even
    # number
    n_els = len(inp_data)

    if n_els % 2:
        inp_time = inp_time[:-1]
        inp_data = inp_data[:-1, :]

        n_els = len(inp_data)

    try:
        num_fields = inp_data.shape[1]
    except IndexError:
        num_fields = 1
        inp_data = inp_data[:, np.newaxis]

    # Set NaN values to zero so FFT works
    idx_nans = np.isnan(inp_data)
    inp_data[idx_nans] = 0.0

    # Bandpass filter field data
    fs = calc_fs(inp)
    fn = fs / 2
    f = np.linspace(-fn, fn, n_els)

    # FFT and remove frequencies
    for nn in range(num_fields):
        inp_temp = np.fft.fft(inp_data[:, nn])
        inp_temp = np.fft.fftshift(inp_temp)

        inp_temp[np.abs(f) < f_min] = 0
        inp_temp[np.abs(f) > f_max] = 0

        inp_temp = np.fft.ifftshift(inp_temp)
        inp_data[:, nn] = np.fft.ifft(inp_temp)

    # Put back original NaNs
    inp_data[idx_nans] = np.nan

    # Return data in the same format as input
    if num_fields == 1:
        out = ts_scalar(inp_time, inp_data, attrs=inp.attrs)
    elif num_fields == 3:
        out = ts_vec_xyz(inp_time, inp_data, attrs=inp.attrs)
    else:
        raise ValueError("Invalid shape")

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
corr_deriv.py

@author : Louis RICHARD
"""

import numpy as np

from astropy.time import Time

from . import find_closest


def corr_deriv(x1=None, x2=None, fla=False):
    """Correlate the derivatives of two time series

    Parameters
    ----------
    x1 : xarray.DataArray
        Time series of the first to variable to correlate with.

    x2 : xarray.DataArray
        Time series of the second to variable to correlate with.

    fla : bool
        Flag if False (default) returns time instants of common highest first and second
        derivatives. If True returns time instants of common highest first derivative and zeros
        crossings.

    Returns
    -------
    t1_d, t2_d : numpy.ndarray
        Time instants of common highest first derivatives.

    t1_dd, t2_dd : numpy.ndarray
        Time instants of common highest second derivatives or zero crossings.

    """

    # 1st derivative
    tx1 = Time(x1.time.data, format="datetime64").unix
    x1 = x1.data
    dtx1 = tx1[:-1] + 0.5 * np.diff(tx1)
    dx1 = np.diff(x1)

    tx2 = Time(x2.time.data, format="datetime64").unix
    x2 = x2.data
    dtx2 = tx2[:-1] + 0.5 * np.diff(tx2)
    dx2 = np.diff(x2)

    ind_zeros1 = np.where(np.sign(dx1[:-1] * dx1[1:]) < 0)[0]
    if ind_zeros1 == 0:
        ind_zeros1 = ind_zeros1[1:]

    ind_zeros2 = np.where(np.sign(dx2[:-1] * dx2[1:]) < 0)[0]
    if ind_zeros2 == 0:
        ind_zeros2 = ind_zeros2[1:]

    ind_zeros1_p = np.where(dx1[ind_zeros1 - 1] - dx1[ind_zeros1] > 0)[0]
    ind_zeros2_p = np.where(dx2[ind_zeros2 - 1] - dx2[ind_zeros2] > 0)[0]

    ind_zeros1_m = np.where(dx1[ind_zeros1 - 1] - dx1[ind_zeros1] < 0)[0]
    ind_zeros2_m = np.where(dx2[ind_zeros2 - 1] - dx2[ind_zeros2] < 0)[0]

    ind1_p = ind_zeros1[ind_zeros1_p]
    ind1_m = ind_zeros1[ind_zeros1_m]

    t_zeros1_p = dtx1[ind1_p] + (dtx1[ind1_p + 1] - dtx1[ind1_p]) / (
                1 + np.abs(dx1[ind1_p + 1]) / np.abs(dx1[ind1_p]))
    t_zeros1_m = dtx1[ind1_m] + (dtx1[ind1_m + 1] - dtx1[ind1_m]) / (
                1 + np.abs(dx1[ind1_m + 1]) / np.abs(dx1[ind1_m]))

    ind2_p = ind_zeros2[ind_zeros2_p]
    ind2_m = ind_zeros2[ind_zeros2_m]

    t_zeros2_p = dtx2[ind2_p] + (dtx2[ind2_p + 1] - dtx2[ind2_p]) / (
                1 + np.abs(dx2[ind2_p + 1]) / np.abs(dx2[ind2_p]))
    t_zeros2_m = dtx2[ind2_m] + (dtx2[ind2_m + 1] - dtx2[ind2_m]) / (
                1 + np.abs(dx2[ind2_m + 1]) / np.abs(dx2[ind2_m]))

    # Remove repeating points
    t_zeros1_p = np.delete(t_zeros1_p, np.where(np.diff(t_zeros1_p) == 0)[0])
    t_zeros2_p = np.delete(t_zeros2_p, np.where(np.diff(t_zeros2_p) == 0)[0])

    # Define identical pairs of two time axis
    t1_d_p, t2_d_p, _, _ = find_closest(t_zeros1_p, t_zeros2_p)
    t1_d_m, t2_d_m, _, _ = find_closest(t_zeros1_m, t_zeros2_m)

    t1_d = np.vstack([t1_d_p, t1_d_m])
    t1_d = t1_d[t1_d[:, 0].argsort(), 0]

    t2_d = np.vstack([t2_d_p, t2_d_m])
    t2_d = t2_d[t2_d[:, 0].argsort(), 0]

    if fla:
        # zero crossings
        ind_zeros1 = np.where(np.sign(x1[:-1] * x1[1:]) < 0)[0]
        ind_zeros2 = np.where(np.sign(x2[:-1] * x2[1:]) < 0)[0]

        ind_zeros1 = np.delete(ind_zeros1, np.where(ind_zeros1 == 1)[0])
        ind_zeros2 = np.delete(ind_zeros2, np.where(ind_zeros2 == 1)[0])

        ind_zeros1_p = np.where(x1[ind_zeros1 - 1] - x1[ind_zeros1] > 0)[0]
        ind_zeros2_p = np.where(x2[ind_zeros2 - 1] - x2[ind_zeros2] > 0)[0]

        ind_zeros1_m = np.where(x1[ind_zeros1 - 1] - x1[ind_zeros1] < 0)[0]
        ind_zeros2_m = np.where(x2[ind_zeros2 - 1] - x2[ind_zeros2] < 0)[0]

        ind1_p = ind_zeros1[ind_zeros1_p]
        ind1_m = ind_zeros1[ind_zeros1_m]

        t_zeros1_p = tx1[ind1_p] + (tx1[ind1_p + 1] - tx1[ind1_p]) / (
                    1 + np.abs(x1[ind1_p + 1]) / np.abs(x1[ind1_p]))
        t_zeros1_m = tx1[ind1_m] + (tx1[ind1_m + 1] - tx1[ind1_m]) / (
                    1 + np.abs(x1[ind1_m + 1]) / np.abs(x1[ind1_m]))

        ind2_p = ind_zeros2[ind_zeros2_p]
        ind2_m = ind_zeros2[ind_zeros2_m]

        t_zeros2_p = tx2[ind2_p] + (tx2[ind2_p + 1] - tx2[ind2_p]) / (
                    1 + np.abs(x2[ind2_p + 1]) / np.abs(x2[ind2_p]))
        t_zeros2_m = tx2[ind2_m] + (tx2[ind2_m + 1] - tx2[ind2_m]) / (
                    1 + np.abs(x2[ind2_m + 1]) / np.abs(x2[ind2_m]))

    else:
        # 2nd derivative
        dd_tx1 = dtx1[:-1] + 0.5 * np.diff(dtx1)
        ddx1 = np.diff(dx1)

        dd_tx2 = dtx2[:-1] + 0.5 * np.diff(dtx2)
        ddx2 = np.diff(dx2)

        ind_zeros1 = np.where(np.sign(ddx1[:-1] * ddx1[1:]) < 0)[0]
        ind_zeros2 = np.where(np.sign(ddx2[:-1] * ddx2[1:]) < 0)[0]

        ind_zeros1 = np.delete(ind_zeros1, np.where(ind_zeros1 == 1)[0])
        ind_zeros2 = np.delete(ind_zeros2, np.where(ind_zeros2 == 1)[0])

        ind_zeros1_p = np.where(ddx1[ind_zeros1 - 1] - ddx1[ind_zeros1] > 0)[0]
        ind_zeros2_p = np.where(ddx2[ind_zeros2 - 1] - ddx2[ind_zeros2] > 0)[0]

        ind_zeros1_m = np.where(ddx1[ind_zeros1 - 1] - ddx1[ind_zeros1] < 0)[0]
        ind_zeros2_m = np.where(ddx2[ind_zeros2 - 1] - ddx2[ind_zeros2] < 0)[0]

        ind1_p = ind_zeros1[ind_zeros1_p]
        ind1_m = ind_zeros1[ind_zeros1_m]

        t_zeros1_p = dd_tx1[ind1_p] + (dd_tx1[ind1_p + 1] - dd_tx1[ind1_p]) / (
                    1 + np.abs(ddx1[ind1_p + 1]) / np.abs(ddx1[ind1_p]))
        t_zeros1_m = dd_tx1[ind1_m] + (dd_tx1[ind1_m + 1] - dd_tx1[ind1_m]) / (
                    1 + np.abs(ddx1[ind1_m + 1]) / np.abs(ddx1[ind1_m]))

        ind2_p = ind_zeros2[ind_zeros2_p]
        ind2_m = ind_zeros2[ind_zeros2_m]

        t_zeros2_p = dd_tx2[ind2_p] + (dd_tx2[ind2_p + 1] - dd_tx2[ind2_p]) / (
                    1 + np.abs(ddx2[ind2_p + 1]) / np.abs(ddx2[ind2_p]))
        t_zeros2_m = dd_tx2[ind2_m] + (dd_tx2[ind2_m + 1] - dd_tx2[ind2_m]) / (
                    1 + np.abs(ddx2[ind2_m + 1]) / np.abs(ddx2[ind2_m]))

    # Define identical pairs of two time axis
    [t1_dd_p, t2_dd_p] = find_closest(t_zeros1_p, t_zeros2_p)
    [t1_dd_m, t2_dd_m] = find_closest(t_zeros1_m, t_zeros2_m)

    t1_dd = np.vstack([t1_dd_p, t1_dd_m])
    t1_dd = t1_dd[t1_dd[:, 0].argsort(), 0]

    t2_dd = np.vstack([t2_dd_p, t2_dd_m])
    t2_dd = t2_dd[t2_dd[:, 0].argsort(), 0]

    return t1_d, t2_d, t1_dd, t2_dd



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
eb_nrf.py

@author : Louis RICHARD
"""

import numpy as np
import xarray as xr

from . import resample, dot, normalize, cross


def eb_nrf(e=None, b=None, v=None, flag=0):
    """Find E and B in MP system given B and MP normal vector.

    Parameters
    ----------
    e : xarray.DataArray
        Time series of the electric field.

    b : xarray.DataArray
        Time series of the magnetic field.

    v : xarray.DataArray
        Normal vector.

    flag : int or numpy.ndarray
        to fill.

    Returns
    -------
    out : DataArray
        to fill.

    """

    assert e is not None and isinstance(e, xr.DataArray)
    assert b is not None and isinstance(b, xr.DataArray)
    assert v is not None and isinstance(v, xr.DataArray)

    if isinstance(flag, int):
        if flag == 1:
            flag_case = "b"
        else:
            flag_case = "a"
        
        l_direction = None

    elif isinstance(flag, np.ndarray) and np.size(flag) == 3:
        l_direction = flag
        flag_case = "c"
    else:
        raise TypeError("Invalid flag type")

    if flag_case == "a":
        be = resample(b, e).data

        nl = be / np.linalg.norm(be, axis=0)[:, None]  # along the B
        nn = np.cross(np.cross(be, v), be)  # closest to given vn vector
        nn = nn / np.linalg.norm(nn)[:, None]
        nm = np.cross(nn, nl)  # in (vn x b) direction

        # estimate e in new coordinates
        en = dot(e, nn)
        el = dot(e, nl)
        em = dot(e, nm)
        emp = np.hstack([el, em, en])

    elif flag_case == "b":
        nn = v / np.linalg.norm(v)
        nm = normalize(np.cross(nn, np.mean(b)))
        nl = cross(nm, nn)

        # estimate e in new coordinates
        en = dot(e, nn)
        el = dot(e, nl)
        em = dot(e, nm)
        emp = np.hstack([el, em, en])

    elif flag_case == "c":
        nn = normalize(v)
        nm = normalize(np.cross(nn, l_direction))
        nl = cross(nm, nn)

        # estimate e in new coordinates
        en = dot(e, nn)
        el = dot(e, nl)
        em = dot(e, nm)

        emp = np.hstack([el, em, en])
    else:
        raise ValueError("Invalid flag_case")

    out = xr.DataArray(e.time.data, emp, e.attrs)

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
calc_feeps_omni.py

@author : Louis RICHARD
"""

import warnings
import numpy as np
import xarray as xr

from .feeps_split_integral_ch import feeps_split_integral_ch
from .feeps_remove_sun import feeps_remove_sun


def calc_feeps_omni(inp_dataset):
    """Computes the omni-directional FEEPS spectrogram from a Dataset that contains the spectrogram
    of all eyes.

    Parameters
    ----------
    inp_dataset : xarray.Dataset
        Dataset with energy spectrum of every eyes.

    Returns
    -------
    out : xarray.DataArray
        OMNI energy spectrum from the input.

    """

    var = inp_dataset.attrs

    if var["dtype"] == "electron":
        energies = np.array(
            [33.2, 51.90, 70.6, 89.4, 107.1, 125.2, 146.5, 171.3, 200.2, 234.0, 273.4, 319.4, 373.2,
             436.0, 509.2])
    else:
        energies = np.array(
            [57.9, 76.8, 95.4, 114.1, 133.0, 153.7, 177.6, 205.1, 236.7, 273.2, 315.4, 363.8, 419.7,
             484.2, 558.6])

    # set unique energy bins per spacecraft; from DLT on 31 Jan 2017
    e_corr = {"e": [14.0, -1.0, -3.0, -3.0], "i": [0.0, 0.0, 0.0, 0.0]}
    g_fact = {"e": [1.0, 1.0, 1.0, 1.0], "i": [0.84, 1.0, 1.0, 1.0]}

    energies += e_corr[var["dtype"][0]][var["mmsId"]-1]

    # percent error around energy bin center to accept data for averaging;
    # anything outside of energies[i] +/- en_chk*energies[i] will be changed
    # to NAN and not averaged
    en_chk = 0.10

    # top_sensors = eyes['top']
    # bot_sensors = eyes['bottom']

    inp_dataset_clean, _ = feeps_split_integral_ch(inp_dataset)

    inp_dataset_clean_sun_removed = feeps_remove_sun(inp_dataset_clean)

    eye_list = list(inp_dataset_clean_sun_removed.keys())
    tmp_data = inp_dataset_clean_sun_removed[eye_list[0]]

    d_all_eyes = np.empty(
        (tmp_data.shape[0], tmp_data.shape[1], len(inp_dataset_clean_sun_removed)))
    d_all_eyes[:] = np.nan

    for i, k in enumerate(eye_list):
        d_all_eyes[..., i] = inp_dataset_clean_sun_removed[k].data

        try:
            diff_en_ch = inp_dataset_clean_sun_removed[k].coords[
                "Differential_energy_channels"].data
            ie = np.where(np.abs(energies - diff_en_ch) > en_chk * energies)

            if ie[0].size != 0:
                d_all_eyes[:, ie[0], i] = np.nan

        except Warning:
            print('NaN in energy table encountered; sensor T{}'.format(k))

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=RuntimeWarning)
        flux_omni = np.nanmean(d_all_eyes, axis=2)

    flux_omni *= g_fact[var["dtype"][0]][var["mmsId"]-1]

    t, attrs = inp_dataset_clean_sun_removed.time.data
    attrs = inp_dataset_clean_sun_removed[eye_list[0]].attrs

    out = xr.DataArray(flux_omni, coords=[t, energies], dims=["time", "energy"], attrs=attrs)

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
struct_func.py

@author : Louis RICHARD
"""

import numpy as np
import xarray as xr


def struct_func(x=None, scale=None, order=2):
    """Returns the structure function of a time series

    .. math::

       y= \\frac{1}{N-s}\\sum_{i=1}^{N-s}(x_i - x_{i+s})^o

    where :math:`s` is the scale, and :math:`o` is the order.

    Parameters
    ----------
    x : xarray.DataArray
        Input time series.

    scale : list or numpy.ndarray
        A list or an array containing the scales to calculate.

    order : int
        Order of the exponential of the structure function.

    Returns
    -------
    values : xarray.DataArray
        An xarray containing the structure functions, one per product in the original time
        series. The index coordinate contains the scale value, and the attribute 'order' keeps a
        record on the order used for its calculation.

    """

    assert x is not None and isinstance(x, xr.DataArray)

    if scale is None:
        scale = [1]

    data = x.data

    if len(data.shape) == 1:
        data = data[:, np.newaxis]
    else:
        pass

    result = []
    for s in scale:
        result.append(np.mean(np.abs((data[s:, :] - data[:-s, :]) ** order), axis=0))

    result = np.array(result)

    c = x.dims[1]
    cols = x.coords[c].data

    result = xr.DataArray(result, coords=[scale, cols], dims=["scale", c], attrs=x.attrs)

    result.attrs['order'] = order

    return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
get_data.py

@author : Louis RICHARD
"""

from astropy.time import Time

from .split_vs import split_vs
from .list_files import list_files
from .get_dist import get_dist
from .get_ts import get_ts
from ..pyrf import ts_append
from ..pyrf import dist_append


def get_data(var_str="", tint=None, mms_id="1", verbose=True):
    """Load a variable. var_str must be in var (see below)

    * EPHEMERIS :
        "R_gse", "R_gsm"

    * FGM :
        "B_gsm_fgm_srvy_l2", "B_gsm_fgm_brst_l2", "B_gse_fgm_srvy_l2",
        "B_gse_fgm_brst_l2", "B_bcs_fgm_srvy_l2", "B_bcs_fgm_brst_l2",
        "B_dmpa_fgm_srvy_l2", "B_dmpa_fgm_brst_l2"

    * DFG & AFG :
        "B_gsm_dfg_srvy_l2pre", "B_gse_dfg_srvy_l2pre", "B_dmpa_dfg_srvy_l2pre",
        "B_bcs_dfg_srvy_l2pre", "B_gsm_afg_srvy_l2pre", "B_gse_afg_srvy_l2pre",
        "B_dmpa_afg_srvy_l2pre", "B_bcs_afg_srvy_l2pre"

    * SCM :
        "B_gse_scm_brst_l2"

    * EDP :
        "Phase_edp_fast_l2a", "Phase_edp_slow_l2a", "Sdev12_edp_slow_l2a",
        "Sdev34_edp_slow_l2a", "Sdev12_edp_fast_l2a", "Sdev34_edp_fast_l2a",
        "E_dsl_edp_brst_l2", "E_dsl_edp_fast_l2", "E_dsl_edp_brst_ql",
        "E_dsl_edp_fast_ql", "E_dsl_edp_slow_l2", "E_gse_edp_brst_l2",
        "E_gse_edp_fast_l2", "E_gse_edp_slow_l2", "E2d_dsl_edp_brst_l2pre",
        "E2d_dsl_edp_fast_l2pre", "E2d_dsl_edp_brst_ql", "E2d_dsl_edp_fast_ql",
        "E2d_dsl_edp_l2pre", "E2d_dsl_edp_fast_l2pre", "E2d_dsl_edp_brst_l2pre",
        "E_dsl_edp_l2pre", "E_dsl_edp_fast_l2pre", "E_dsl_edp_brst_l2pre",
        "E_dsl_edp_slow_l2pre", "E_ssc_edp_brst_l2a", "E_ssc_edp_fast_l2a",
        "E_ssc_edp_slow_l2a", "V_edp_fast_sitl", "V_edp_slow_sitl",
        "V_edp_slow_l2", "V_edp_fast_l2", "V_edp_brst_l2"

    * FPI Ions :
        "Vi_dbcs_fpi_brst_l2", "Vi_dbcs_fpi_fast_l2", "Vi_dbcs_fpi_l2",
        "Vi_gse_fpi_ql", "Vi_gse_fpi_fast_ql", "Vi_dbcs_fpi_fast_ql",
        "Vi_gse_fpi_fast_l2", "Vi_gse_fpi_brst_l2", "partVi_gse_fpi_brst_l2",
        "Ni_fpi_brst_l2", "partNi_fpi_brst_l2", "Ni_fpi_brst",
        "Ni_fpi_fast_l2", "Ni_fpi_ql", "DEFi_fpi_fast_ql",
        "DEFi_fpi_fast_l2", "Tperpi_fpi_brst_l2", "Tparai_fpi_brst_l2",
        "partTperpi_fpi_brst_l2", "partTparai_fpi_brst_l2", "Ti_dbcs_fpi_brst_l2",
        "Ti_dbcs_fpi_brst", "Ti_dbcs_fpi_fast_l2", "Ti_gse_fpi_ql",
        "Ti_dbcs_fpi_ql", "Ti_gse_fpi_brst_l2", "Pi_dbcs_fpi_brst_l2",
        "Pi_dbcs_fpi_brst", "Pi_dbcs_fpi_fast_l2", "Pi_gse_fpi_ql",
        "Pi_gse_fpi_brst_l2"

    * FPI Electrons :
        "Ve_dbcs_fpi_brst_l2", "Ve_dbcs_fpi_brst", "Ve_dbcs_fpi_ql",
        "Ve_dbcs_fpi_fast_l2", "Ve_gse_fpi_ql", "Ve_gse_fpi_fast_l2",
        "Ve_gse_fpi_brst_l2", "partVe_gse_fpi_brst_l2", "DEFe_fpi_fast_ql",
        "DEFe_fpi_fast_l2", "Ne_fpi_brst_l2", "partNe_fpi_brst_l2",
        "Ne_fpi_brst", "Ne_fpi_fast_l2", "Ne_fpi_ql",
        "Tperpe_fpi_brst_l2", "Tparae_fpi_brst_l2", "partTperpe_fpi_brst_l2",
        "partTparae_fpi_brst_l2", "Te_dbcs_fpi_brst_l2", "Te_dbcs_fpi_brst",
        "Te_dbcs_fpi_fast_l2", "Te_gse_fpi_ql", "Te_dbcs_fpi_ql",
        "Te_gse_fpi_brst_l2", "Pe_dbcs_fpi_brst_l2", "Pe_dbcs_fpi_brst",
        "Pe_dbcs_fpi_fast_l2", "Pe_gse_fpi_ql", "Pe_gse_fpi_brst_l2",

    * HPCA :
        "Nhplus_hpca_srvy_l2", "Nheplus_hpca_srvy_l2", "Nheplusplus_hpca_srvy_l2",
        "Noplus_hpca_srvy_l2", "Tshplus_hpca_srvy_l2", "Tsheplus_hpca_srvy_l2",
        "Tsheplusplus_hpca_srvy_l2", "Tsoplus_hpca_srvy_l2", "Vhplus_dbcs_hpca_srvy_l2",
        "Vheplus_dbcs_hpca_srvy_l2", "Vheplusplus_dbcs_hpca_srvy_l2", "Voplus_dbcs_hpca_srvy_l2",
        "Phplus_dbcs_hpca_srvy_l2", "Pheplus_dbcs_hpca_srvy_l2", "Pheplusplus_dbcs_hpca_srvy_l2",
        "Poplus_dbcs_hpca_srvy_l2", "Thplus_dbcs_hpca_srvy_l2", "Theplus_dbcs_hpca_srvy_l2",
        "Theplusplus_dbcs_hpca_srvy_l2", "Toplus_dbcs_hpca_srvy_l2", "Vhplus_gsm_hpca_srvy_l2",
        "Vheplus_gsm_hpca_srvy_l2", "Vheplusplus_gsm_hpca_srvy_l2", "Voplus_gsm_hpca_srvy_l2",
        "Nhplus_hpca_brst_l2", "Nheplus_hpca_brst_l2", "Nheplusplus_hpca_brst_l2",
        "Noplus_hpca_brst_l2", "Tshplus_hpca_brst_l2", "Tsheplus_hpca_brst_l2",
        "Tsheplusplus_hpca_brst_l2", "Tsoplus_hpca_brst_l2", "Vhplus_dbcs_hpca_brst_l2",
        "Vheplus_dbcs_hpca_brst_l2", "Vheplusplus_dbcs_hpca_brst_l2", "Voplus_dbcs_hpca_brst_l2",
        "Phplus_dbcs_hpca_brst_l2", "Pheplus_dbcs_hpca_brst_l2", "Pheplusplus_dbcs_hpca_brst_l2",
        "Poplus_dbcs_hpca_brst_l2", "Thplus_dbcs_hpca_brst_l2", "Theplus_dbcs_hpca_brst_l2",
        "Theplusplus_dbcs_hpca_brst_l2", "Toplus_dbcs_hpca_brst_l2", "Vhplus_gsm_hpca_brst_l2",
        "Vheplus_gsm_hpca_brst_l2", "Vheplusplus_gsm_hpca_brst_l2", "Voplus_gsm_hpca_brst_l2",
        "Phplus_gsm_hpca_brst_l2", "Pheplus_gsm_hpca_brst_l2", "Pheplusplus_gsm_hpca_brst_l2",
        "Poplus_gsm_hpca_brst_l2", "Thplus_gsm_hpca_brst_l2", "Theplus_gsm_hpca_brst_l2",
        "Theplusplus_gsm_hpca_brst_l2", "Toplus_gsm_hpca_brst_l2"

    Parameters
    ----------
    var_str : str
        Key of the target variable (see above).

    tint : list of str
        Time interval.

    mms_id : str or int
        Index of the target spacecraft.

    verbose : bool
        Set to True to follow the loading. Default is True.

    Returns
    -------
    out : xarray.DataArray
        Time series of the target variable of measured by the target spacecraft over the selected
        time interval.

    Examples
    --------
    >>> from pyrfu import mms
    >>> # Define time interval
    >>> tint = ["2019-09-14T07:54:00.000", "2019-09-14T08:11:00.000"]
    >>> # Index of mms spacecraft
    >>> mms_id = 1
    >>> # Load magnetic field from FGM
    >>> b_xyz = mms.get_data("B_gse_fgm_brst_l2", tint, mms_id)

    """

    if not var_str:
        raise ValueError("get_data requires at least 2 arguments")

    if tint is None:
        raise ValueError("get_data requires at least 2 arguments")

    if not isinstance(mms_id, str):
        mms_id = str(mms_id)

    # Translate short names to names readable for splitVs
    if var_str == "dfg_ql_srvy":
        var_str = "B_dmpa_dfg_srvy_ql"

    elif var_str == "afg_ql_srvy":
        var_str = "B_dmpa_afg_srvy_ql"

    elif var_str == "Nhplus_hpca_sitl":
        var_str = "Nhplus_hpca_srvy_sitl"

    elif var_str in ["R_gse", "R_gsm", "V_gse", "V_gsm"]:
        var_str = "_".join([var_str, "mec", "srvy", "l2"])

    var = split_vs(var_str)

    mms_id_str = "mms{}".format(mms_id)

    var["dtype"] = None

    vdf_flag = False

    if var["inst"] == "mec":
        cdf_name = "_".join([mms_id_str, "mec", var["param"].lower(), var["cs"]])

        var["dtype"] = "epht89d"

    elif var["inst"] == "fsm":
        cdf_name = "_".join([mms_id_str, var["inst"], "b", var["cs"], var["tmmode"], var["lev"]])

        var["dtype"] = "8khz"

    elif var["inst"] in ["fgm", "dfg", "afg"]:
        if var["lev"] == "l2":
            cdf_name = "_".join(
                [mms_id_str, var["inst"], "b", var["cs"], var["tmmode"], var["lev"]])

        elif var["lev"] == "l2pre":
            cdf_name = "_".join(
                [mms_id_str, var["inst"], "b", var["cs"], var["tmmode"], var["lev"]])

        elif var["lev"] == "ql":
            if var["cs"] == "dmpa":
                cdf_name = "_".join([mms_id_str, var["inst"], var["tmmode"], var["cs"]])
            elif var["cs"] == "gsm":
                cdf_name = "_".join([mms_id_str, var["inst"], var["tmmode"], var["cs"], "dmpa"])
            else:
                raise ValueError("Invalid coordinates")
        else:
            raise InterruptedError("Should not be here")

    elif var["inst"] == "fpi":
        # get specie
        if var["param"][-1] == "i":
            sensor = "dis"

        elif var["param"][-1] == "e":
            sensor = "des"

        else:
            raise ValueError("invalid specie")

        if var["lev"] in ["l2", "l2pre", "l1b"]:
            if var["param"][:2].lower() == "pd":
                vdf_flag = True

                var["dtype"] = "{}-dist".format(sensor)

            else:
                if len(var["param"]) > 4:
                    if var["param"][:4] == "part":
                        var["dtype"] = "{}-partmoms".format(sensor)

                    else:
                        var["dtype"] = "{}-moms".format(sensor)

                else:
                    var["dtype"] = "{}-moms".format(sensor)

        elif var["lev"] == "ql":
            var["dtype"] = sensor

        else:
            raise InterruptedError("Should not be here")

        if var["param"].lower() in ["pdi", "pde", "pderri", "pderre"]:
            if var["param"][:-1].lower() == "pd":
                cdf_name = "_".join([mms_id_str, sensor, "dist", var["tmmode"]])

            elif len(var["param"]) == 6 and var["param"][:-1].lower() == "pderr":
                cdf_name = "_".join([mms_id_str, sensor, "disterr", var["tmmode"]])

            else:
                raise InterruptedError("Should not be here")

        # Number density
        elif var["param"].lower() in ["ne", "ni"]:
            if var["lev"] in ["l2", "l2pre"]:
                cdf_name = "_".join([mms_id_str, sensor, "numberdensity", var["tmmode"]])

            elif var["lev"] == "l1b":
                cdf_name = "_".join([mms_id_str, sensor, "numberdensity"])

            elif var["lev"] == "ql":
                cdf_name = "_".join([mms_id_str, sensor, "numberdensity", "fast"])

            elif var["lev"] == "sitl":
                cdf_name = "_".join([mms_id_str, "fpi", sensor.upper(), "numberDensity"])

            else:
                raise InterruptedError("Should not be here")

        # Number density
        elif var["param"].lower() in ["nbge", "nbgi"]:
            if var["lev"] in ["l2", "l2pre"]:
                cdf_name = "_".join([mms_id_str, sensor, "numberdensity_bg", var["tmmode"]])

            elif var["lev"] == "l1b":
                cdf_name = "_".join([mms_id_str, sensor, "numberdensity_bg"])

            elif var["lev"] == "ql":
                cdf_name = "_".join([mms_id_str, sensor, "numberdensity_bg", "fast"])

            else:
                raise InterruptedError("Should not be here")

        # Partial number density
        elif var["param"].lower() in ["partni", "partne"]:
            if var["lev"] == "l2":
                # only for l2 data
                cdf_name = "_".join([mms_id_str, sensor, "numberdensity", "part", var["tmmode"]])

            else:
                raise InterruptedError("Should not be here")

        # Energy flux omni
        elif var["param"].lower() in ["defi", "defe"]:
            if var["lev"] == "ql":
                cdf_name = "_".join([mms_id_str, sensor, "energyspectr", "omni", "fast"])

            elif var["lev"] == "l2":
                cdf_name = "_".join([mms_id_str, sensor, "energyspectr", "omni", var["tmmode"]])

            else:
                raise InterruptedError("Should not be here")

        # Energy flux omni
        elif var["param"].lower() in ["defbgi", "defbge"]:
            if var["lev"] == "l2":
                cdf_name = "_".join([mms_id_str, sensor, "spectr_bg", var["tmmode"]])

            else:
                raise InterruptedError("Should not be here")

        # Energies
        elif var["param"].lower() in ["energyi", "energye"]:
            if var["lev"] == "ql":
                cdf_name = "_".join([mms_id_str, sensor, "energy", "fast"])

            elif var["lev"] == "l2":
                cdf_name = "_".join([mms_id_str, sensor, "energy", var["tmmode"]])

            else:
                raise InterruptedError("Should not be here")

        # Parallel and perpendicular temperatures
        elif var["param"].lower() in ["tparai", "tparae", "tperpi", "tperpe"]:
            # Field Aligned component (either para or perp)
            tmp_fac = var["param"][1:5]

            if var["lev"] == "l2":
                cdf_name = "_".join([mms_id_str, sensor, "temp{}".format(tmp_fac), var["tmmode"]])

            else:
                raise InterruptedError("Should not be here")

        # Partial moments parallel and perpandiculat temperatures
        elif var["param"].lower() in ["parttparai", "parttparae", "parttperpi", "parttperpe"]:
            # Field Aligned component (either para or perp)
            tmp_fac = var["param"][5:9]

            if var["lev"] == "l2":
                cdf_name = "_".join(
                    [mms_id_str, sensor, "temp{}".format(tmp_fac), "part", var["tmmode"]])

            else:
                raise InterruptedError("Should not be here")

        # Temperature and pressure tensors
        elif var["param"].lower() in ["ti", "te", "pi", "pe"]:
            if var["param"][0].lower() == "t":
                mom_type = "temptensor"  # temperature

            elif var["param"][0].lower() == "p":
                mom_type = "prestensor"  # pressure

            else:
                raise InterruptedError("Should not be here")

            cdf_name = "_".join([mms_id_str, sensor, mom_type, var["cs"], var["tmmode"]])

        elif var["param"].lower() in ["pbgi", "pbge"]:
            mom_type = "pres_bg"

            cdf_name = "_".join([mms_id_str, sensor, mom_type, var["tmmode"]])

        # Partial temperature and pressure tensors
        elif var["param"].lower() in ["partti", "partte", "partpi", "partpe"]:
            if var["param"][4] == "T":
                mom_type = "temptensor"  # temperature

            elif var["param"][4] == "P":
                mom_type = "prestensor"  # pressure

            else:
                raise InterruptedError("Should not be here")

            cdf_name = "_".join([mms_id_str, sensor, mom_type, "part", var["cs"], var["tmmode"]])

        # spintone
        elif var["param"].lower() in ["sti", "ste"]:
            if var["lev"] in ["l2", "l2pre"]:
                cdf_name = "_".join(
                    [mms_id_str, sensor, "bulkv_spintone", var["cs"], var["tmmode"]])

            elif var["lev"] == "ql":
                cdf_name = "_".join(
                    [mms_id_str, sensor, "bulkv_spintone", var["cs"], var["tmmode"]])

            else:
                raise InterruptedError("Should not be here")

        # Bulk velocities
        elif var["param"].lower() in ["vi", "ve"]:
            if var["lev"] in ["l2", "l2pre"]:
                cdf_name = "_".join([mms_id_str, sensor, "bulkv", var["cs"], var["tmmode"]])

            elif var["lev"] == "ql":
                cdf_name = "_".join([mms_id_str, sensor, "bulkv", var["cs"], var["tmmode"]])

            else:
                raise InterruptedError("Should not be here")

        # Error bulk velocities
        elif var["param"].lower() in ["errvi", "errve"]:
            if var["lev"] == "l2":
                cdf_name = "_".join([mms_id_str, sensor, "bulkv", "err", var["tmmode"]])

            else:
                raise InterruptedError("Only l2 partmoms available now")

        # Partial bulk velocities
        elif var["param"].lower() in ["partvi", "partve"]:
            if var["lev"] == "l2":
                cdf_name = "_".join([mms_id_str, sensor, "bulkv", "part", var["cs"], var["tmmode"]])

            else:
                raise InterruptedError("Only l2 partmoms available now")

        # Heat flux
        elif var["param"].lower() in ["qi", "qe"]:
            if var["lev"] in ["l2", "l2pre"]:
                cdf_name = "_".join([mms_id_str, sensor, "heatq", var["cs"], var["tmmode"]])

            elif var["lev"] == "ql":
                cdf_name = "_".join([mms_id_str, sensor, "heatq", var["cs"], var["tmmode"]])

            else:
                raise InterruptedError("Should not be here")

        # Heat flux error
        elif var["param"].lower() in ["errqi", "errqe"]:
            if var["lev"] in ["l2"]:
                cdf_name = "_".join([mms_id_str, sensor, "heatq", "err", var["tmmode"]])
            else:
                raise InterruptedError("Should not be here")

        elif var["param"].lower() in ["padlowene", "padmidene", "padhighene"]:
            # Energy range
            en_range = var["param"][3:-1]

            if var["lev"] == "l2":
                cdf_name = "_".join([mms_id_str, sensor, "pitchangdist", en_range, var["tmmode"]])

            else:
                raise InterruptedError("Only l2 partmoms available now")

        else:
            raise InterruptedError("Should not be here")

    # Hot Plasma Composition Analyser
    elif var["inst"] == "hpca":
        param, var["dtype"] = [var["param"][0], "moments"]

        # Get specie name
        ion = var["param"][1:]

        if ion[0] == "s":
            param, ion = [param+ion[0], ion[1:]]

        assert ion in ["hplus", "heplus", "heplusplus", "oplus"]

        # Number density
        if "N" in var["param"]:
            cdf_name = "_".join([mms_id_str, "hpca", ion, "number_density"])

        # Bulk velocity
        elif "V" in var["param"]:
            cdf_name = "_".join([mms_id_str, "hpca", ion, "ion_bulk_velocity"])

        # Scalar temperature
        elif "Ts" in var["param"]:
            cdf_name = "_".join([mms_id_str, "hpca", ion, "scalar_temperature"])

        # Pressure tensor
        elif "P" in var["param"]:
            cdf_name = "_".join([mms_id_str, "hpca", ion, "ion_pressure"])

        # Temperature tensor
        elif "T" in var["param"]:
            cdf_name = "_".join([mms_id_str, "hpca", ion, "temperature_tensor"])

        else:
            raise ValueError("Unrecognized param")

        # Tensor (vector or matrix) add coordinate system to cdf_name
        if var["to"] > 0:
            if var["cs"].lower() == "gsm":
                cdf_name = "_".join([cdf_name, "GSM"])

            elif var["cs"].lower() == "dbcs":
                pass

            else:
                raise ValueError("invalid CS")

    # Search Coil Magnetometer
    elif var["inst"] == "scm":
        if var["lev"] != "l2":
            raise InterruptedError("not implemented yet")

        var["dtype"] = "scb"

        cdf_name = "_".join([mms_id_str, "scm", "acb", var["cs"], "scb", var["tmmode"], var["lev"]])

    elif var["inst"] == "dsp":
        if var["lev"] == "l2":
            if var["param"][0].lower() == "e":
                var["dtype"] = "{}psd".format(var["param"][0].lower())

                cdf_name = "_".join([mms_id_str, "dsp", var["dtype"], "omni"])

            elif var["param"][0].lower() == "b":
                var["dtype"] = "{}psd".format(var["param"][0].lower())

                cdf_name = "_".join(
                    [mms_id_str, "dsp", var["dtype"], "omni", var["tmmode"], var["lev"]])

            else:
                raise ValueError("Should not be here")

    # Spin-plane Double mmsId instrument
    elif var["inst"] == "edp":
        if var["lev"] == "sitl":
            if var["param"].lower() == "e":
                param = "_".join(["dce_xyz", var["cs"]])

                var["dtype"] = "dce"

            elif var["param"].lower() == "e2d":
                param = "_".join(["dce_xyz", var["cs"]])

                var["dtype"] = "dce2d"

            elif var["param"].lowr() == "v":
                param = "scpot"

                var["dtype"] = "scpot"

            else:
                raise ValueError("Invalid param")

            cdf_name = "_".join([mms_id_str, "edp", param, var["tmmode"], var["lev"]])

        elif var["lev"] == "ql":
            if var["param"].lower() == "e":
                param = "_".join(["dce_xyz", var["cs"]])

                var["dtype"] = "dce"

            elif var["param"].lower() == "e2d":
                param = "_".join(["dce_xyz", var["cs"]])

                var["dtype"] = "dce2d"

            else:
                raise ValueError("Invalid param")

            cdf_name = "_".join([mms_id_str, "edp", param])

        elif var["lev"] == "l1b":
            if var["param"].lower() == "e":
                param = "dce_sensor"

                var["dtype"] = "dce"

            elif var["param"].lower() == "v":
                param = "dcv_sensor"

                var["dtype"] = "dce"

            else:
                raise ValueError("Invalid param")

            cdf_name = "_".join([mms_id_str, "edp", param])

        elif var["lev"] == "l2a":
            var["dtype"] = "dce2d"

            if var["param"].lower() == "phase":
                param = "phase"

            elif var["param"].lower() in ["es12", "es34"]:
                param = "espin_p{}".format(var["param"][2:4])

            elif var["param"].lower() == "adcoff":
                param = "adc_offset"

            elif var["param"].lower() in ["sdev12", "sdev34"]:
                param = "sdevfit_p{}".format(var["param"][4:6])

            elif var["param"].lower() == "e":
                param = "dce"

            else:
                raise ValueError("Invalid param")

            cdf_name = "_".join([mms_id_str, "edp", param, var["tmmode"], var["lev"]])

        else:
            if var["param"].lower() == "e":
                param = "dce_{}".format(var["cs"])

                var["dtype"] = "dce"

            elif var["param"].lower() == "epar":
                param = "dce_par_epar"

                var["dtype"] = "dce"

            elif var["param"].lower() == "e2d":
                param = "dce_{}".format(var["cs"])

                var["dtype"] = "dce2d"

            elif var["param"].lower() == "v":
                param = "scpot"

                var["dtype"] = "scpot"

            elif var["param"].lower() == "v6":
                var["dtype"] = "scpot"

                param = "dcv"

            else:
                raise ValueError("Invalid param")

            cdf_name = "_".join([mms_id_str, "edp", param, var["tmmode"], var["lev"]])

    else:
        raise ValueError("not implemented yet")

    files = list_files(tint, mms_id, var)

    if not files:
        raise ValueError("No files found. Make sure that the data_path is correct")

    if verbose:
        print(f"Loading {cdf_name}...")

    out = None

    for file in files:
        if vdf_flag:
            out = dist_append(out, get_dist(file, cdf_name, tint))

        else:
            out = ts_append(out, get_ts(file, cdf_name, tint))

    if out.time.data.dtype == "float64":
        out.time.data = Time(1e-9 * out.time.data, format="unix").datetime

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
read_feeps_sector_mask_csv.py

@author : Louis RICHARD
"""

import os
import csv
import numpy as np

from astropy.time import Time


def read_feeps_sector_masks_csv(tint=None):
    """
    Reads the FEEPS sectors to mask due to sunlight contamination from csv files.x

    Parameters
    ----------
    tint : list of str
        time range of interest [starttime, endtime] with the format "YYYY-MM-DD", "YYYY-MM-DD"]
        or to specify more or less than a day ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

    Returns
    -------
    mask : dict
        Hash table containing the sectors to mask for each spacecraft and sensor ID

    """
    masks = {}

    dates = [1447200000.0000000,  # 11/11/2015
             1468022400.0000000,  # 7/9/2016
             1477612800.0000000,  # 10/28/2016
             1496188800.0000000,  # 5/31/2017
             1506988800.0000000,  # 10/3/2017
             1538697600.0000000]  # 10/5/2018

    # find the file closest to the start time
    nearest_date = dates[(np.abs(np.array(dates)-Time(tint[0], format="isot").unix)).argmin()]
    str_date = Time(nearest_date, format="unix").datetime.strftime("%Y%m%d")

    for mms_sc in np.arange(1, 5):
        csv_file = os.sep.join([os.path.dirname(os.path.abspath(__file__)), "sun",
                                "MMS{:d}_FEEPS_ContaminatedSectors_{}.csv".format(mms_sc,
                                                                                  str_date)])

        csv_file = open(csv_file, 'r')

        csv_reader = csv.reader(csv_file)

        csv_data = []

        for line in csv_reader:
            csv_data.append([float(l) for l in line])

        csv_file.close()

        csv_data = np.array(csv_data)

        for i in range(0, 12):
            mask_vals = []
            for val_idx in range(len(csv_data[:, i])):
                if csv_data[val_idx, i] == 1:
                    mask_vals.append(val_idx)

            masks["mms{:d}_imask_top-{:d}".format(mms_sc, i + 1)] = mask_vals

        for i in range(0, 12):
            mask_vals = []

            for val_idx in range(len(csv_data[:, i+12])):
                if csv_data[val_idx, i+12] == 1:
                    mask_vals.append(val_idx)

            masks["mms{:d}_imask_bottom-{:d}".format(mms_sc, i + 1)] = mask_vals

    return masks



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
calc_feeps_pad.py

@author : Louis RICHARD
"""

import warnings
import numpy as np
import xarray as xr

from astropy.time import Time

from .feeps_pitch_angles import feeps_pitch_angles
from .get_feeps_active_eyes import get_feeps_active_eyes


def calc_feeps_pad(inp_dataset=None, b_bcs=None, bin_size=16.3636, energy=None):
    """Compute pitch angle distribution using FEEPS data.

    Parameters
    ----------
    inp_dataset : xarray.Dataset
        Energy spectrum of all eyes.

    b_bcs : xarray.DataArray
        Time series of the magnetic field in spacecraft coordinates.

    bin_size : float
        Width of the pitch angles bins.

    energy : list of float
        Energy range of particles.

    Returns
    -------
    pad : xarray.DataArray
        Time series of the pitch angle distribution.

    """

    assert inp_dataset is not None and isinstance(inp_dataset, xr.Dataset)
    assert b_bcs is not None and isinstance(b_bcs, xr.DataArray)
    assert isinstance(bin_size, float)
    assert energy is None or isinstance(energy, list)

    if energy is None:
        energy = [70, 600]

    var = inp_dataset.attrs
    mms_id = var["mmsId"]

    if var["dtype"] == "electron":
        dangresp = 21.4  # deg
    elif var["dtype"] == "ion":
        dangresp = 10.0  # deg
    else:
        raise ValueError("Invalid specie")

    if energy[0] < 32.0:
        raise ValueError("Please select a starting energy of 32 keV or above")

    n_pabins = 180 / bin_size
    pa_bins = [180. * pa_bin / n_pabins for pa_bin in range(0, int(n_pabins)+1)]
    pa_label = [180. * pa_bin / n_pabins + bin_size / 2. for pa_bin in range(0, int(n_pabins))]

    pitch_angles, idx_maps = feeps_pitch_angles(inp_dataset, b_bcs)
    pa_times = pitch_angles.time
    pa_data = pitch_angles.data

    trange = Time(np.hstack([pa_times.data.min(), pa_times.data.max()]), format="datetime64").isot

    eyes = get_feeps_active_eyes(var, trange, mms_id)

    pa_data_map = {}

    if var["tmmode"] == "srvy":
        if var["dtype"] == "electron":
            pa_data_map["top-electron"] = idx_maps["electron-top"]
            pa_data_map["bottom-electron"] = idx_maps["electron-bottom"]
        elif var["dtype"] == "ion":
            pa_data_map["top-ion"] = idx_maps["electron-top"]
            pa_data_map["bottom-ion"] = idx_maps["ion-bottom"]
        else:
            raise ValueError("Invalid specie")

    elif var["tmmode"] == "brst":
        # note: the following are indices of the top/bottom sensors in pa_data
        # they should be consistent with pa_dlimits.labels
        pa_data_map["top-electron"] = np.arange(9)
        pa_data_map["bottom-electron"] = np.arange(9, 18)

        # and ions:
        pa_data_map["top-ion"], pa_data_map["bottom-ion"] = [[0, 1, 2], [3, 4, 5]]

    sensor_types = ["top", "bottom"]

    if var["dtype"] == "electron":
        n_times = len(pa_times)
        n_top, n_bottom = [len(pa_data_map["top-electron"]), len(pa_data_map["bot-electron"])]

        dflux, dpa = [np.zeros([n_times, n_top + n_bottom]) for _ in range(2)]

    elif var["dtype"] == "ion":
        n_times = len(pa_times)
        n_top, n_bottom = [len(pa_data_map["top-ion"]), len(pa_data_map["bot-ion"])]

        dflux, dpa = [np.zeros([n_times, n_top + n_bottom]) for _ in range(2)]
    else:
        raise TypeError("Invalid specie")

    for s_type in sensor_types:
        pa_map = pa_data_map[s_type + "-" + var["dtype"]]

        particle_idxs = [eye-1 for eye in eyes[s_type]]

        for isen, sensor_num in enumerate(particle_idxs):
            var_name = "{}-{:d}".format(s_type, sensor_num + 1)

            t, data, = inp_dataset[var_name].time.data, inp_dataset[var_name].data
            energies = inp_dataset[var_name].Differential_energy_channels.data

            # remove any 0s before averaging
            data[data == 0] = "nan"

            # assumes all energies are NaNs if the first is
            if np.isnan(energies[0]):
                continue

            # energy indices to use:
            idx = np.where((energies >= energy[0]) & (energies <= energy[1]))

            with warnings.catch_warnings():
                warnings.simplefilter("ignore", category=RuntimeWarning)
                dflux[:, pa_map[isen]] = np.nanmean(data[:, idx[0]], axis=1)

            dpa[:, pa_map[isen]] = pa_data[:, pa_map[isen]]

    # we need to replace the 0.0s left in after populating dpa with NaNs; these
    # 0.0s are left in there because these points aren't covered by sensors loaded
    # for this datatype/data_rate
    dpa[dpa == 0] = "nan"

    pa_flux, delta_pa = [np.zeros([len(pa_times), int(n_pabins)]), (pa_bins[1] - pa_bins[0]) / 2.0]

    # Now loop through PA bins and time, find the telescopes where there is data in those bins and
    # average it up!
    for pa_idx in range(len(pa_times)):
        for ipa in range(0, int(n_pabins)):
            if not np.isnan(dpa[pa_idx, :][0]):
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore", category=RuntimeWarning)
                    ind = np.where((dpa[pa_idx, :] + dangresp >= pa_label[ipa]-delta_pa)
                                & (dpa[pa_idx, :]-dangresp < pa_label[ipa]+delta_pa))

                    if ind[0].size != 0:
                        if len(ind[0]) > 1:
                            pa_flux[pa_idx, ipa] = np.nanmean(dflux[pa_idx, ind[0]], axis=0)
                        else:
                            pa_flux[pa_idx, ipa] = dflux[pa_idx, ind[0]]

    pa_flux[pa_flux == 0] = "nan"  # fill any missed bins with NAN

    pad = xr.DataArray(pa_flux, coords=[t, pa_label], dims=["time", "pa"], attrs=var)

    return pad



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
calc_disprel_tm.py

@author : Louis RICHARD
"""

import numpy as np
import xarray as xr
from scipy import optimize


def calc_disprel_tm(v=None, v_err=None, tau=None, tau_err=None):
    """Computes dispersion relation from velocities and period given by the timing method.

    Parameters
    ----------
    v : xarray.DataArray
        Time series of the velocities.

    v_err : xarray.DataArray
        Time series of the error on velocities.

    tau : xarray.DataArray
        Time series of the periods.

    tau_err : xarray.DataArray
        Time series of the error on period.

    Returns
    -------
    out : xarray.Dataset
        DataSet containing the frequency, the wavelength, the wavenumber. Also includes the
        errors and the fit (e.g Vph phase velocity).

    See also
    --------
    pyrfu.pyrf.c_4_v : Calculates velocity or time shift of discontinuity.

    """

    assert v is not None and isinstance(v, xr.DataArray)
    assert v_err is not None and isinstance(v_err, xr.DataArray)
    assert tau is not None and isinstance(tau, xr.DataArray)
    assert tau_err is not None and isinstance(tau_err, xr.DataArray)

    # Frequency, wavelength, wave number
    omega, lamb, k = [2 * np.pi / tau.data, v * tau.data, 2 * np.pi/(v * tau.data)]

    # Estimate propagation of the errors
    # Error on frequency
    omega_err = omega*((tau_err / tau) / (1 + tau_err / tau))

    # Error on wavelength
    lamb_err = v_err * tau

    # Error on wave number
    k_err = k*((lamb_err/lamb)/(1+lamb_err/lamb))

    def model_tau_v(x, a):
        return a / x

    fit_tau_v, cov_tau_v = optimize.curve_fit(model_tau_v, tau, v, 1,
                                              sigma=np.sqrt(v_err ** 2 + tau_err ** 2))
    sigma_tau_v = np.sqrt(np.diagonal(cov_tau_v))

    # High resolution prediction
    hires_tau = np.logspace(np.log10(5), np.log10(2e3), int(1e4))
    predict_v = model_tau_v(hires_tau, *fit_tau_v)

    # 95% confidence interval
    bound_upper_v = model_tau_v(hires_tau, *(fit_tau_v + 1.96*sigma_tau_v))
    bound_lower_v = model_tau_v(hires_tau, *(fit_tau_v - 1.96*sigma_tau_v))

    def model_k_w(x, a):
        return a * x

    fit, cov = optimize.curve_fit(model_k_w, k, omega, 1,
                                  sigma=np.sqrt(omega_err ** 2 + k_err ** 2))
    sigma_k_w = np.sqrt(np.diagonal(cov))

    # High resolution prediction
    hires_k = np.linspace(0, 0.003, int(1e4))
    predict_w = model_k_w(hires_k, *fit)

    # 95% confidence interval
    bound_upper_w = model_k_w(hires_k, *(fit + 1.96 * sigma_k_w))
    bound_lower_w = model_k_w(hires_k, *(fit - 1.96 * sigma_k_w))

    out_dict = {'tau': tau, 'tau_err': (["tau"], tau_err), 'v': (["tau"], v),
                'v_err': (["tau"], v_err),
                'lamb': (["tau"], lamb),
                'lamb_err': (["tau"], lamb_err), 'k': k, 'k_err': (["k"], k_err),
                'omega': (["k"], omega),
                'omega_err': (["k"], omega_err), 'hires_k': hires_k,
                'pred_omega': (["hires_k"], predict_w),
                'bound_upper': (["hires_k"], bound_upper_w),
                'bound_lower': (["hires_k"], bound_lower_w),
                'hires_tau': hires_tau, 'pred_v': (["hires_tau"], predict_v),
                'bound_upper_v': (["hires_tau"], bound_upper_v),
                'bound_lower_v': (["hires_tau"], bound_lower_v),
                'l': fit_tau_v, 'vph': fit, 'sigma_k_w': sigma_k_w}

    out = xr.Dataset(out_dict)

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
time_clip.py

@author : Louis RICHARD
"""

import bisect
import datetime
import numpy as np
import xarray as xr

from dateutil import parser
from astropy.time import Time


def time_clip(inp=None, tint=None):
    """Time clip the input (if time interval is TSeries clip between start and stop).

    Parameters
    ----------
    inp : xarray.DataArray
        Time series of the quantity to clip.

    tint : xarray.DataArray or numpy.ndarray or list
        Time interval can be a time series, a array of datetime64 or a list.

    Returns
    -------
    out : xarray.DataArray
        Time series of the time clipped input.

    """

    assert inp is not None and isinstance(inp, xr.DataArray)
    assert tint is not None and isinstance(tint, (list, xr.DataArray, np.ndarray))

    if isinstance(tint, xr.DataArray):
        t_start, t_stop = [tint.time.data[0], tint.time.data[-1]]

    elif isinstance(tint, np.ndarray):
        if isinstance(tint[0], datetime.datetime) and isinstance(tint[-1], datetime.datetime):
            t_start, t_stop = [tint.time[0], tint.time[-1]]

        else:
            raise TypeError('Values must be in Datetime64')

    else:
        t_start, t_stop = [parser.parse(tint[0]), parser.parse(tint[-1])]

    idx_min = bisect.bisect_left(inp.time.data, Time(t_start, format="datetime").datetime64)
    idx_max = bisect.bisect_right(inp.time.data, Time(t_stop, format="datetime").datetime64)

    coord = [inp.time.data[idx_min:idx_max]]

    if len(inp.coords) > 1:
        for k in inp.dims[1:]:
            coord.append(inp.coords[k].data)

    out = xr.DataArray(inp.data[idx_min:idx_max, ...], coords=coord, dims=inp.dims, attrs=inp.attrs)
    out.time.attrs = inp.time.attrs

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
filt.py

@author : Louis RICHARD
"""

import xarray as xr
import numpy as np
from scipy import signal


# noinspection PyTupleAssignmentBalance
def filt(inp=None, f_min=0, f_max=1, n=-1):
    """Filters input quantity.

    Parameters
    ----------
    inp : xarray.DataArray
        Time series of the variable to filter.

    f_min : float
        Lower limit of the frequency range.

    f_max : float
        Upper limit of the frequency range.

    n : int
        Order of the elliptic filter.

    Returns
    -------
    out : xarray.DataArray
        Time series of the filtered signal.

    Examples
    --------
    >>> from pyrfu import mms, pyrf
    >>> # Time interval
    >>> tint = ["2017-07-18T13:03:34.000", "2017-07-18T13:07:00.000"]
    >>> # Spacecraft index
    >>> mms_id = 1
    >>> # Load magnetic and electric fields
    >>> b_xyz = mms.get_data("B_gse_fgm_brst_l2", tint, mms_id)
    >>> e_xyz = mms.get_data("E_gse_edp_brst_l2", tint, mms_id)
    >>> # Convert E to field aligned coordinates
    >>> e_xyzfac = pyrf.convert_fac(e_xyz, b_xyz, [1,0,0])
    >>> # Bandpass filter E waveform
    >>> e_xyzfac_hf = pyrf.filt(e_xyzfac, 4, 0, 3)
    >>> e_xyzfac_lf = pyrf.filt(e_xyzfac, 0, 4, 3)

    """

    assert inp is not None and isinstance(inp, xr.DataArray)

    fs = 1 / (np.median(np.diff(inp.time)).astype(int) * 1e-9)

    # Data of the input
    inp_data = inp.data

    f_min, f_max = [f_min / (fs / 2), f_max / (fs / 2)]

    if f_max > 1:
        f_max = 1

    # Parameters of the elliptic filter. fact defines the width between stopband and passband
    r_p, r_s, fact = [0.5, 60, 1.1]

    if f_min == 0:
        b1, a1 = [None] * 2
        b2, a2 = [None] * 2

        if n == -1:
            n, f_max = signal.ellipord(f_max, np.min([f_max * fact, 0.9999]), r_p, r_s)

        b, a = signal.ellip(n, r_p, r_s, f_max, btype="lowpass")
    elif f_max == 0:
        b1, a1 = [None] * 2
        b2, a2 = [None] * 2

        if n == -1:
            n, f_min = signal.ellipord(f_min, np.min([f_min * fact, 0.9999]), r_p, r_s)

        b, a = signal.ellip(n, r_p, r_s, f_min, btype="highpass")
    else:
        b, a = [None] * 2

        if n == -1:
            n, f_max = signal.ellipord(f_max, np.min([f_max * 1.3, 0.9999]), r_p, r_s)

        b1, a1 = signal.ellip(n, r_p, r_s, f_max)

        if n == -1:
            n, f_min = signal.ellipord(f_min, f_min * .75, r_p, r_s)

        b2, a2 = signal.ellip(n, r_p, r_s, f_min)

    try:
        n_c = inp_data.shape[1]
    except IndexError:
        n_c = 1
        inp_data = inp_data[:, np.newaxis]

    out_data = np.zeros(inp_data.shape)

    if f_min != 0 and f_max != 0:
        for i_col in range(n_c):
            out_data[:, i_col] = signal.filtfilt(b1, a1, inp_data[:, i_col])
            out_data[:, i_col] = signal.filtfilt(b2, a2, out_data[:, i_col])
    else:
        for i_col in range(n_c):
            out_data[:, i_col] = signal.filtfilt(b, a, inp_data[:, i_col])

    if n_c == 1:
        out_data = out_data[:, 0]

    out = xr.DataArray(out_data, coords=inp.coords, dims=inp.dims, attrs=inp.attrs)

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
pid_4sc.py

@author : Louis RICHARD
"""

import numpy as np
import xarray as xr

from ..mms import rotate_tensor

from .c_4_grad import c_4_grad
from .avg_4sc import avg_4sc
from .trace import trace
from .ts_tensor_xyz import ts_tensor_xyz
from .ts_scalar import ts_scalar


def pid_4sc(r_mms=None, v_mms=None, p_mms=None, b_mms=None):
    """
    Compute Pi-D term using definition of [10]_ as :

    .. math::

        \\textrm{Pi}-\\textrm{D} = - \\mathbf{\\Pi}_{ij}\\mathbf{D}_{ij}

    with :math:`\\mathbf{\\Pi}_{ij}` the deviatoric part of the pressure tensor :

    .. math::

        \\mathbf{\\Pi}_{ij} = \\mathbf{P}_{ij} - p\\delta_{ij}

        p = \\frac{1}{3}\\mathbf{P}_{ii}


    and :math:`\\mathbf{D}_{ij}` the deviatoric part of the strain tensor :

    .. math::

        \\mathbf{D}_{ij} =
        \\frac{1}{2}\\left ( \\partial_i \\mathbf{u}_j + \\partial_j \\mathbf{u}_i\\right )
        - \\frac{1}{3}\\theta\\delta_{ij}

        \\theta = \\nabla . \\mathbf{u}


    Parameters
    ----------
    r_mms : list of xarray.DataArray
        Time series of the position of the 4 spacecraft.

    v_mms : list of xarray.DataArray
        Time series of the bulk velocities of the 4 spacecraft.

    p_mms : list of xarray.DataArray
        Time series of the pressure tensor of the 4 spacecraft.

    b_mms : list of xarray.DataArray
        Time series of the background magnetic field of the 4 spacecraft.

    Returns
    -------
    pid : xarray.DataArray
        Time series of the Pi-D.

    References
    ----------
    .. [10]  Yang, Y., Matthaeus, W. H., Parashar, T. N., Wu, P., Wan, M., Shi, Y.,
            et al. (2017). Energy transfer channels and turbulence cascade in Vlasov-Maxwell
            turbulence. Physical Review E, 95, 061201.
            doi : https://doi.org/10.1103/PhysRevE.95.061201

    """

    # Check input
    assert r_mms is not None and isinstance(r_mms, list) and isinstance(r_mms[0], xr.DataArray)
    assert v_mms is not None and isinstance(v_mms, list) and isinstance(v_mms[0], xr.DataArray)
    assert p_mms is not None and isinstance(p_mms, list) and isinstance(p_mms[0], xr.DataArray)
    assert b_mms is not None and isinstance(b_mms, list) and isinstance(b_mms[0], xr.DataArray)

    time = v_mms[0].time.data

    # Compute pressure tensor and background magnetic field at the center of mass of the tetrahedron
    p_xyz = avg_4sc(p_mms)
    b_xyz = avg_4sc(b_mms)

    # Compute divergence and gradient tensor of the bulk velocity. Yang's notation
    theta, grad_u = [c_4_grad(r_mms, v_mms, method) for method in ["div", "grad"]]

    # Define identity tensor
    identity_3d = np.zeros((len(grad_u), 3, 3))
    np.einsum("jii->ji", identity_3d)[:] = 1

    # strain tensor
    eps_xyz = (grad_u.data + np.transpose(grad_u.data, [0, 2, 1])) / 2

    # Deviatoric part of the strain tensor
    d_xyz = eps_xyz - theta.data[:, np.newaxis, np.newaxis] * identity_3d / 2
    d_xyz = ts_tensor_xyz(time, d_xyz)

    # Convert tensors to field aligned coordinates
    d_fac, p_fac = [rotate_tensor(tensor, "fac", b_xyz, "pp") for tensor in [d_xyz, p_xyz]]

    # Isotropic scalar pressure
    p = trace(p_fac) / 3

    # Deviatoric part of the pressure tensor
    pi_fac = p_fac.data - p.data[:, np.newaxis, np.newaxis] * identity_3d
    pi_fac = ts_tensor_xyz(time, pi_fac)

    # Compute Pi-D
    pid = np.sum(np.sum(pi_fac.data * d_fac.data, axis=1), axis=1)
    pid = ts_scalar(time, pid)

    # Flatten tensors
    idx_flat = [0, 4, -1, 1, 2, 5]
    d_flat, pi_flat = [np.reshape(tensor.data, [len(tensor), 9]) for tensor in [d_fac, pi_fac]]
    d_flat, pi_flat = [tensor[:, idx_flat] for tensor in [d_flat, pi_flat]]
    # Compute components of the double contraction sum
    d_coeff = d_flat * pi_flat

    times = p_xyz.time.data
    index = np.arange(1, 7)
    d_coeff = xr.DataArray(d_coeff, coords=[times, index], dims=["time", "index"])

    return pid, d_coeff



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
lowpass.py

@author : Louis RICHARD
"""

import xarray as xr

from scipy import signal


def lowpass(inp=None, f_cut=None, fhz=None):
    """Filter the data through low or highpass filter with max frequency f_cut and subtract from
    the original.

    Parameters
    ----------
    inp : xarray.DataArray
        Time series of the input variable.

    f_cut : float
        Cutoff frequency.

    fhz : float
        Sampling frequency.

    Returns
    -------
    out : xarray.DataArray
        Time series of the filter data.

    """

    assert inp is not None and isinstance(inp, xr.DataArray)
    assert f_cut is not None and isinstance(f_cut, float)
    assert fhz is not None and isinstance(fhz, float)

    data = inp.data

    # Remove trend
    data_detrend = signal.detrend(data, type='linear')
    rest = data - data_detrend

    # Elliptic filter
    f_nyq, rp, rs, n = [fhz / 2, 0.1, 60, 4]

    [b, a] = signal.ellip(n, rp, rs, f_cut / f_nyq, output='ba')

    # Filter data
    out_data = signal.filtfilt(b, a, data_detrend) + rest

    out = xr.DataArray(out_data, coords=inp.coords, dims=inp.dims)

    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
rotate_tensor.py

@author : Louis RICHARD
"""

import numpy as np

from pyrfu.pyrf.resample import resample
from pyrfu.pyrf.ts_tensor_xyz import ts_tensor_xyz


def rotate_tensor(*args):
    """Rotates pressure or temperature tensor into another coordinate system.

    Parameters
    ----------
    PeIJ/Peall : xarray.DataArray
        Time series of either separated terms of the tensor or the complete tensor.
        If columns (PeXX,PeXY,PeXZ,PeYY,PeYZ,PeZZ)

    flag : str
        Flag of the target coordinates system :
            * "fac" : 	Field-aligned coordinates, requires background magnetic field Bback,
                        optional flag "pp" :math:`\\mathbf{P}_{\\perp 1} = \\mathbf{P}_{\\perp
                        2}` or "qq" :math:`\\mathbf{P}_{\\perp 1}` and
                        :math:`\\mathbf{P}_{\\perp 2}` are most unequal, sets P23 to zero.

            * "rot" :	Arbitrary coordinate system, requires new x-direction xnew, new y and z
                        directions ynew, znew (if not included y and z directions are orthogonal
                        to xnew and closest to the original y and z directions)

            * "gse" : GSE coordinates, requires MMS spacecraft number 1--4 MMSnum

    Returns
    -------
    Pe : xarray.DataArray
        Time series of the pressure or temperature tensor in field-aligned, user-defined,
        or GSE coordinates. For "fac" Pe = [Ppar P12 P13; P12 Pperp1 P23; P13 P23 Pperp2].
        For "rot" and "gse" Pe = [Pxx Pxy Pxz; Pxy Pyy Pyz; Pxz Pyz Pzz]

    Examples
    --------
    >>> from pyrfu import mms, pyrf
    >>> # Time interval
    >>> tint = ["2015-10-30T05:15:20.000", "2015-10-30T05:16:20.000"]
    >>> # Spacecraft index
    >>> mms_id = 1
    >>> # Load magnetic field and ion temperature tensor
    >>> b_xyz = mms.get_data("B_gse_fgm_srvy_l2", tint, mms_id)
    >>> t_xyz_i = mms.get_data("Ti_gse_fpi_fast_l2", tint, mms_id)
    >>> # Compute ion temperature in field aligned coordinates
    >>> t_xyzfac_i = mms.rotate_tensor(t_xyz_i, "fac", b_xyz, "pp")

    TODO : change input
    """

    nargin = len(args)

    # Check input and load pressure/temperature terms
    if isinstance(args[1], str):
        rot_flag = args[1]
        rot_flag_pos = 1
        p_all = args[0]
        p_times = p_all.time.data

        if p_all.data.ndim == 3:
            p_tensor = p_all
        else:
            p_tensor = np.reshape(p_all.data, (p_all.shape[0], 3, 3))
            p_tensor = ts_tensor_xyz(p_times, p_tensor)
    else:
        raise SystemError("critical','Something is wrong with the input.")

    ppeq, qqeq = [0, 0]

    rot_mat = np.zeros((len(p_times), 3, 3))

    if rot_flag[0] == "f":
        print("notice : Transforming tensor into field-aligned coordinates.")

        if nargin == rot_flag_pos:
            raise ValueError("B TSeries is missing.")

        b_back = args[rot_flag_pos + 1]
        b_back = resample(b_back, p_tensor)

        if nargin == 4:
            if isinstance(args[3], str) and args[3][0] == "p":
                ppeq = 1
            elif isinstance(args[3], str) and args[3][0] == "q":
                qqeq = 1
            else:
                raise ValueError("Flag not recognized no additional rotations applied.")

        if nargin == 9:
            if isinstance(args[8], str) and args[8][0] == "p":
                ppeq = 1
            elif isinstance(args[8], str) and args[8][0] == "q":
                qqeq = 1
            else:
                raise ValueError("Flag not recognized no additional rotations applied.")

        b_vec = b_back / np.linalg.norm(b_back, axis=1, keepdims=True)

        r_x = b_vec.data
        r_y = np.array([1, 0, 0])
        r_z = np.cross(r_x, r_y)
        r_z /= np.linalg.norm(r_z, axis=1, keepdims=True)
        r_y = np.cross(r_z, r_x)
        r_y /= np.linalg.norm(r_y, axis=1, keepdims=True)

        rot_mat[:, 0, :], rot_mat[:, 1, :], rot_mat[:, 2, :] = [r_x, r_y, r_z]

    elif rot_flag[0] == "r":
        print("notice : Transforming tensor into user defined coordinate system.")

        if nargin == rot_flag_pos:
            raise ValueError("Vector(s) is(are) missing.")

        vectors = list(args[rot_flag_pos + 1:])

        if len(vectors) == 1:
            r_x = vectors[0]

            if len(r_x) != 3:
                raise TypeError("Vector format not recognized.")

            r_x /= np.linalg.norm(r_x, keepdims=True)
            r_y = np.array([0, 1, 0])
            r_z = np.cross(r_x, r_y)
            r_z /= np.linalg.norm(r_z, keepdims=True)
            r_y = np.cross(r_z, r_x)
            r_y /= np.linalg.norm(r_y, keepdims=True)

        elif len(vectors) == 3:
            r_x, r_y, r_z = [r / np.linalg.norm(r, keepdims=True) for r in vectors]
            # TODO : add check that vectors are orthogonal
        else:
            raise TypeError("Vector format not recognized.")

        rot_mat[:, 0, :] = np.ones((len(p_times), 1)) * r_x
        rot_mat[:, 1, :] = np.ones((len(p_times), 1)) * r_y
        rot_mat[:, 2, :] = np.ones((len(p_times), 1)) * r_z

    else:
        raise ValueError("Flag is not recognized.")

    p_tensor_p = np.zeros((len(p_times), 3, 3))

    for ii in range(len(p_times)):
        rot_temp = np.squeeze(rot_mat[ii, :, :])

        p_tensor_p[ii, :, :] = np.matmul(np.matmul(rot_temp, np.squeeze(p_tensor.data[ii, :, :])),
                                         np.transpose(rot_temp))

    if ppeq:
        print("notice : Rotating tensor so perpendicular diagonal components are equal.")
        thetas = 0.5 * np.arctan(
            (p_tensor_p[:, 2, 2] - p_tensor_p[:, 1, 1]) / (2 * p_tensor_p[:, 1, 2]))

        for ii, theta in enumerate(thetas):
            if np.isnan(theta):
                theta = 0

            rot_temp = np.array(
                [[1, 0, 0], [0, np.cos(theta), np.sin(theta)], [0, -np.sin(theta), np.cos(theta)]])

            p_tensor_p[ii, :, :] = np.matmul(
                np.matmul(rot_temp, np.squeeze(p_tensor_p[ii, :, :])), np.transpose(rot_temp))

    if qqeq:
        print("notice : Rotating tensor so perpendicular diagonal components are most unequal.")
        thetas = 0.5 * np.arctan(
            (2 * p_tensor_p[:, 1, 2]) / (p_tensor_p[:, 2, 2] - p_tensor_p[:, 1, 1]))

        for ii, theta in enumerate(thetas):
            rot_temp = np.array(
                [[1, 0, 0], [0, np.cos(theta), -np.sin(theta)], [0, np.sin(theta), np.cos(theta)]])

            p_tensor_p[ii, :, :] = np.matmul(
                np.matmul(rot_temp, np.squeeze(p_tensor_p[ii, :, :])), np.transpose(rot_temp))

    # Construct output
    p = ts_tensor_xyz(p_times, p_tensor_p)

    try:
        p.attrs["units"] = args[0].attrs["units"]
    except KeyError:
        pass

    return p
